"""Wrapper to expose Layer 1 native Cython tests to pytest."""

from __future__ import annotations

import sys
import os

# Go up to tests/ directory to find the compiled .so files
# level/cython/test_level_wrapper.py -> level/cython -> level -> advanced -> orderbook -> tests
test_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "..", "..", "..")
)
if test_dir not in sys.path:
    sys.path.insert(0, test_dir)

try:
    import cython_test_level as _native
except ImportError as e:
    import pytest

    pytest.skip(f"Native Cython test module not built: {e}", allow_module_level=True)


# Layer 1: Primitives (27 test functions)


def test_create_orderbook_level_basic():
    _native.test_create_orderbook_level_basic()


def test_create_orderbook_level_default_norders():
    _native.test_create_orderbook_level_default_norders()


def test_create_orderbook_level_zero_price():
    _native.test_create_orderbook_level_zero_price()


def test_create_orderbook_level_zero_size():
    _native.test_create_orderbook_level_zero_size()


def test_create_orderbook_level_large_values():
    _native.test_create_orderbook_level_large_values()


def test_create_orderbook_level_with_ticks_and_lots_basic():
    _native.test_create_orderbook_level_with_ticks_and_lots_basic()


def test_create_orderbook_level_with_ticks_and_lots_default_norders():
    _native.test_create_orderbook_level_with_ticks_and_lots_default_norders()


def test_create_orderbook_level_ticks_rounding():
    _native.test_create_orderbook_level_ticks_rounding()


def test_create_orderbook_level_very_small_tick():
    _native.test_create_orderbook_level_very_small_tick()


def test_create_orderbook_level_large_tick():
    _native.test_create_orderbook_level_large_tick()


def test_convert_price_to_tick_basic():
    _native.test_convert_price_to_tick_basic()


def test_convert_price_from_tick_basic():
    _native.test_convert_price_from_tick_basic()


def test_tick_conversion_roundtrip():
    _native.test_tick_conversion_roundtrip()


def test_convert_size_to_lot_basic():
    _native.test_convert_size_to_lot_basic()


def test_convert_size_from_lot_basic():
    _native.test_convert_size_from_lot_basic()


def test_lot_conversion_roundtrip():
    _native.test_lot_conversion_roundtrip()


def test_convert_zero_price():
    _native.test_convert_zero_price()


def test_convert_zero_size():
    _native.test_convert_zero_size()


def test_swap_levels():
    _native.test_swap_levels()


def test_reverse_levels_basic():
    _native.test_reverse_levels_basic()


def test_reverse_levels_single():
    _native.test_reverse_levels_single()


def test_reverse_levels_even():
    _native.test_reverse_levels_even()


def test_sort_levels_ascending():
    _native.test_sort_levels_ascending()


def test_sort_levels_descending():
    _native.test_sort_levels_descending()


def test_sort_levels_already_sorted():
    _native.test_sort_levels_already_sorted()


def test_create_orderbook_levels():
    _native.test_create_orderbook_levels()


def test_free_orderbook_levels():
    _native.test_free_orderbook_levels()



"""Sub-bucket management declarations."""
from libc.stdint cimport int64_t as i64

from .types cimport EventTokenState


cdef i64 active_sub_index(EventTokenState* state, i64 now)
cdef void refresh_sub_bucket(EventTokenState* state, i64 now, i64 idx)
cdef void reset_state(EventTokenState* state, object config, i64 now)



"""Type stubs for enums.pyx - Orderbook sortedness enums."""

from __future__ import annotations

from enum import IntEnum

class PyOrderbookSortedness(IntEnum):
    """Python-facing enum for specifying orderbook level sortedness.

    Use this enum when constructing PyAdvancedOrderbook to indicate
    how incoming snapshot and delta data is sorted.
    """

    UNKNOWN = 0
    """Sortedness is unknown; levels will be sorted on ingestion."""

    ASCENDING = 1
    """Levels are sorted in ascending order (lowest price first)."""

    DESCENDING = 2
    """Levels are sorted in descending order (highest price first)."""

    BIDS_ASCENDING_ASKS_DESCENDING = 3
    """Bids are ascending, asks are descending."""

    BIDS_DESCENDING_ASKS_ASCENDING = 4
    """Bids are descending, asks are ascending (most common exchange format)."""

class CyOrderbookSortedness(IntEnum):
    """Cython-facing enum for specifying orderbook level sortedness.

    This is a cpdef enum that can be used from both Python and Cython.
    For Python code, prefer using PyOrderbookSortedness.
    """

    UNKNOWN = 0
    ASCENDING = 1
    DESCENDING = 2
    BIDS_ASCENDING_ASKS_DESCENDING = 3
    BIDS_DESCENDING_ASKS_ASCENDING = 4

def py_to_cy_orderbook_sortedness(
    sortedness: PyOrderbookSortedness,
) -> CyOrderbookSortedness:
    """Convert a Python-facing sortedness enum to a Cython-facing sortedness enum.

    Args:
        sortedness: Python sortedness enum value

    Returns:
        Corresponding Cython sortedness enum value
    """
    ...



from libc.stdint cimport uint64_t as u64

from .core cimport CoreAdvancedOrderbook
from .level.level cimport OrderbookLevel, OrderbookLevels
from .enum.enums cimport CyOrderbookSortedness


cdef class AdvancedOrderbook:
    cdef CoreAdvancedOrderbook _core

    cdef void clear(self)
    cdef void consume_snapshot(self, OrderbookLevels asks, OrderbookLevels bids)
    cdef void consume_deltas(self, OrderbookLevels asks, OrderbookLevels bids)
    cdef void consume_bbo(self, OrderbookLevel ask, OrderbookLevel bid)

    cdef double get_mid_price(self)
    cdef double get_bbo_spread(self)
    cdef double get_wmid_price(self)
    cdef double get_volume_weighted_mid_price(self, double size, bint is_base_currency)
    cdef double get_price_impact(self, double size, bint is_buy, bint is_base_currency)
    cdef double get_size_for_price_impact_bps(self, double impact_bps, bint is_buy, bint is_base_currency)
    cdef bint is_bbo_crossed(self, double other_bid_price, double other_ask_price)
    cdef bint does_bbo_price_change(self, double bid_price, double ask_price)




from libc.stdint cimport uint64_t as u64


cdef extern from *:
    unsigned long long __atomic_load_n(unsigned long long* ptr, int memorder)
    void __atomic_store_n(unsigned long long* ptr, unsigned long long val, int memorder)
    unsigned long long __atomic_add_fetch(unsigned long long* ptr, unsigned long long val, int memorder)
    unsigned long long __atomic_sub_fetch(unsigned long long* ptr, unsigned long long val, int memorder)


cdef int _ATOMIC_RELAXED
cdef int _ATOMIC_ACQUIRE
cdef int _ATOMIC_RELEASE
cdef int _ATOMIC_ACQ_REL

cdef u64 atomic_load_acquire(u64* p) nogil
cdef void atomic_store_release(u64* p, u64 v) nogil
cdef u64 atomic_add(u64* p, u64 delta) nogil
cdef u64 atomic_sub(u64* p, u64 delta) nogil



"""
Comprehensive boundary condition tests for AdvancedOrderbook.

Tests edge cases including empty orderbooks, max capacity, pathological data,
precision/rounding, and overflow protection. These tests verify the robustness
of recent bug fixes and ensure correct behavior at system boundaries.
"""

from __future__ import annotations

import pytest

from mm_toolbox.orderbook.advanced import (
    AdvancedOrderbook,
    OrderbookLevel,
    OrderbookLevels,
    PyOrderbookSortedness,
)
from tests.orderbook.advanced.conftest import (
    TICK_SIZE,
    LOT_SIZE,
    _mk_book,
    _make_levels,
    _empty_levels,
    _empty_bid_levels,
)


@pytest.mark.boundary
class TestEmptyOrderbookBoundaries:
    """Test empty orderbook operations and edge cases."""

    def test_empty_orderbook_initialization(self):
        """Verify that num_levels<16 raises ValueError."""
        with pytest.raises(ValueError, match="expected >=64"):
            AdvancedOrderbook(
                tick_size=TICK_SIZE,
                lot_size=LOT_SIZE,
                num_levels=0,
                delta_sortedness=PyOrderbookSortedness.UNKNOWN,
                snapshot_sortedness=PyOrderbookSortedness.UNKNOWN,
            )

    def test_operations_on_never_populated_orderbook(self):
        """Empty arrays returned when orderbook never populated."""
        book = _mk_book(num_levels=64)

        # get_*_numpy() should return empty arrays
        bids = book.get_bids_numpy()
        asks = book.get_asks_numpy()

        assert len(bids) == 0
        assert len(asks) == 0

        # get_mid_price() and get_bbo_spread() should raise on empty orderbook
        with pytest.raises(RuntimeError, match="Empty view"):
            book.get_mid_price()

        with pytest.raises(RuntimeError, match="Empty view"):
            book.get_bbo_spread()

    def test_snapshot_with_zero_levels(self):
        """Snapshot with deletion markers retains zero-size levels."""
        book = _mk_book(num_levels=64)

        # Create snapshots with only deletion markers (size=0, norders=0)
        empty_asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.0], [0.0], [0], TICK_SIZE, LOT_SIZE
        )
        empty_bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.99], [0.0], [0], TICK_SIZE, LOT_SIZE
        )

        book.consume_snapshot(empty_asks, empty_bids)

        # Orderbook retains zero-size levels from snapshots
        bids = book.get_bids_numpy()
        asks = book.get_asks_numpy()
        assert len(bids) == 1
        assert len(asks) == 1
        assert bids["size"][0] == 0.0
        assert asks["size"][0] == 0.0

    def test_successive_deletions_to_empty_state(self):
        """Delete all bids/asks one-by-one until empty."""
        book = _mk_book(num_levels=64)

        # Initialize with 3 levels
        asks, _ = _make_levels(
            [100.0, 100.01, 100.02], [1.0, 1.0, 1.0], with_precision=True
        )
        bids, _ = _make_levels(
            [99.99, 99.98, 99.97], [1.0, 1.0, 1.0], with_precision=True
        )
        book.consume_snapshot(asks, bids)

        # Delete all asks
        delete_ask1 = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.0], [0.0], [0], TICK_SIZE, LOT_SIZE
        )
        delete_ask2 = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.01], [0.0], [0], TICK_SIZE, LOT_SIZE
        )
        delete_ask3 = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.02], [0.0], [0], TICK_SIZE, LOT_SIZE
        )

        book.consume_deltas(delete_ask1, _empty_bid_levels())
        book.consume_deltas(delete_ask2, _empty_bid_levels())
        book.consume_deltas(delete_ask3, _empty_bid_levels())

        # Orderbook should be empty after all deletions
        asks_arr = book.get_asks_numpy()
        assert len(asks_arr) == 0

    def test_ask_delete_then_insert_same_delta(self):
        """Delete the only ask then insert a new ask in the same delta."""
        book = _mk_book(num_levels=64)

        asks, _ = _make_levels([100.0], [1.0], with_precision=True)
        bids, _ = _make_levels([99.90], [1.0], with_precision=True)
        book.consume_snapshot(asks, bids)

        delta_asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.0, 100.02], [0.0, 1.0], [0, 1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(delta_asks, _empty_bid_levels())

        asks_arr = book.get_asks_numpy()
        assert len(asks_arr) == 1
        assert asks_arr["price"][0] == pytest.approx(100.02)

    def test_bid_delete_then_insert_same_delta(self):
        """Delete the only bid then insert a new bid in the same delta."""
        book = _mk_book(num_levels=64)

        asks, _ = _make_levels([100.0], [1.0], with_precision=True)
        bids, _ = _make_levels([99.90], [1.0], with_precision=True)
        book.consume_snapshot(asks, bids)

        delta_bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.90, 99.97], [0.0, 1.0], [0, 1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(_empty_levels(), delta_bids)

        bids_arr = book.get_bids_numpy()
        assert len(bids_arr) == 1
        assert bids_arr["price"][0] == pytest.approx(99.97)

    def test_bbo_deletion_leaving_empty_side(self):
        """One-sided empty after BBO deletion."""
        book = _mk_book(num_levels=64)

        # Initialize with single level on each side
        asks, _ = _make_levels([100.0], [1.0], with_precision=True)
        bids, _ = _make_levels([99.90], [1.0], with_precision=True)
        book.consume_snapshot(asks, bids)

        # Delete the only ask level
        delete_ask = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.0], [0.0], [0], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(delete_ask, _empty_bid_levels())

        # Orderbook should be empty
        asks_arr = book.get_asks_numpy()
        assert len(asks_arr) == 0

    def test_cross_removal_ignored_without_replacement(self):
        """Crossing ask delta without bid replacements is ignored."""
        book = _mk_book(num_levels=64)

        # Initialize with tight bid/ask levels
        asks, _ = _make_levels([100.0, 100.01], [1.0, 1.0], with_precision=True)
        bids, _ = _make_levels([99.99, 99.98], [1.0, 1.0], with_precision=True)
        book.consume_snapshot(asks, bids)

        # Send crossed update: ask below best bid
        crossed_ask = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.95], [1.0], [1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(crossed_ask, _empty_bid_levels())

        bids_arr = book.get_bids_numpy()
        asks_arr = book.get_asks_numpy()
        assert bids_arr["price"][0] == pytest.approx(99.99)
        assert asks_arr["price"][0] == pytest.approx(100.0)

    def test_crossing_bid_delta_ignored_without_replacement(self):
        """Crossing bid delta without ask replacements is ignored."""
        book = _mk_book(num_levels=64)

        asks, _ = _make_levels([100.01, 100.02], [1.0, 1.0], with_precision=True)
        bids, _ = _make_levels([100.00, 99.99], [1.0, 1.0], with_precision=True)
        book.consume_snapshot(asks, bids)

        crossed_bid = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.05], [1.0], [1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(_empty_levels(), crossed_bid)

        bids_arr = book.get_bids_numpy()
        asks_arr = book.get_asks_numpy()
        assert bids_arr["price"][0] == pytest.approx(100.00)
        assert asks_arr["price"][0] == pytest.approx(100.01)

    def test_crossing_ask_delta_with_bid_replacement_applies(self):
        """Crossing ask delta with bid replacements updates both sides."""
        book = _mk_book(num_levels=64)

        asks, _ = _make_levels([100.01, 100.02], [1.0, 1.0], with_precision=True)
        bids, _ = _make_levels([100.00, 99.99], [1.0, 1.0], with_precision=True)
        book.consume_snapshot(asks, bids)

        crossed_ask = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.95], [1.0], [1], TICK_SIZE, LOT_SIZE
        )
        replacement_bid = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.90], [1.0], [1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(crossed_ask, replacement_bid)

        bids_arr = book.get_bids_numpy()
        asks_arr = book.get_asks_numpy()
        assert bids_arr["price"][0] == pytest.approx(99.90)
        assert asks_arr["price"][0] == pytest.approx(99.95)


@pytest.mark.boundary
class TestDeltaBatchTopLevelRemovals:
    """Boundary coverage for delta batches that remove multiple top levels."""

    def test_ask_batch_removals_do_not_duplicate_levels(self):
        """Ensure ask-side removals in one delta do not duplicate prices.

        Returns:
            None
        """
        book = _mk_book(num_levels=64)

        asks, _ = _make_levels(
            [100.00, 100.01, 100.02], [1.0, 1.0, 1.0], with_precision=True
        )
        bids, _ = _make_levels(
            [99.99, 99.98, 99.97], [1.0, 1.0, 1.0], with_precision=True
        )
        book.consume_snapshot(asks, bids)

        delta_asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.00, 100.01, 100.02], [0.0, 0.0, 2.0], [0, 0, 1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(delta_asks, _empty_bid_levels())

        asks_arr = book.get_asks_numpy()
        assert len(asks_arr) == 1
        assert asks_arr["price"][0] == pytest.approx(100.02)
        assert asks_arr["size"][0] == pytest.approx(2.0)
        assert len(asks_arr["ticks"]) == len(set(asks_arr["ticks"]))

    def test_bid_batch_removals_do_not_duplicate_levels(self):
        """Ensure bid-side removals in one delta do not duplicate prices.

        Returns:
            None
        """
        book = _mk_book(num_levels=64)

        asks, _ = _make_levels(
            [100.01, 100.02, 100.03], [1.0, 1.0, 1.0], with_precision=True
        )
        bids, _ = _make_levels(
            [99.99, 99.98, 99.97], [1.0, 1.0, 1.0], with_precision=True
        )
        book.consume_snapshot(asks, bids)

        delta_bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.99, 99.98, 99.97], [0.0, 0.0, 2.0], [0, 0, 1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(_empty_levels(), delta_bids)

        bids_arr = book.get_bids_numpy()
        assert len(bids_arr) == 1
        assert bids_arr["price"][0] == pytest.approx(99.97)
        assert bids_arr["size"][0] == pytest.approx(2.0)
        assert len(bids_arr["ticks"]) == len(set(bids_arr["ticks"]))


@pytest.mark.boundary
class TestMaxCapacityBoundaries:
    """Test max capacity edge cases and overflow protection."""

    def test_initialization_at_max_capacity(self):
        """Create with num_levels=64777216 (ORDERBOOK_MAX_LEVELS)."""
        # This test is slow due to large memory allocation
        # Skip in normal test runs
        pytest.skip("Skipping max capacity initialization test (very slow)")

        book = AdvancedOrderbook(
            tick_size=TICK_SIZE,
            lot_size=LOT_SIZE,
            num_levels=64777216,  # 2^24
            delta_sortedness=PyOrderbookSortedness.UNKNOWN,
            snapshot_sortedness=PyOrderbookSortedness.UNKNOWN,
        )
        assert book is not None

    def test_initialization_above_max_capacity(self):
        """Verify ValueError for overflow (tested at ladder level)."""
        # Note: Validation happens at OrderbookLadder level (ladder.pyx:58)
        # ORDERBOOK_MAX_LEVELS = 16777216 (2^24), so 16777217 should fail instantly
        with pytest.raises(ValueError, match="Invalid max_levels"):
            AdvancedOrderbook(
                tick_size=TICK_SIZE,
                lot_size=LOT_SIZE,
                num_levels=64777217,  # 2^24 + 1, exceeds max
                delta_sortedness=PyOrderbookSortedness.UNKNOWN,
                snapshot_sortedness=PyOrderbookSortedness.UNKNOWN,
            )

    @pytest.mark.slow
    def test_snapshot_at_max_capacity(self, max_capacity_orderbook):
        """Full capacity snapshot insertion."""
        book = max_capacity_orderbook

        # Create 1000 levels
        ask_prices = [100.0 + i * 0.01 for i in range(1000)]
        bid_prices = [99.99 - i * 0.01 for i in range(1000)]
        sizes = [1.0] * 1000
        norders = [1] * 1000

        asks = OrderbookLevels.from_list_with_ticks_and_lots(
            ask_prices, sizes, norders, TICK_SIZE, LOT_SIZE
        )
        bids = OrderbookLevels.from_list_with_ticks_and_lots(
            bid_prices, sizes, norders, TICK_SIZE, LOT_SIZE
        )

        book.consume_snapshot(asks, bids)

        asks_arr = book.get_asks_numpy()
        bids_arr = book.get_bids_numpy()

        assert len(asks_arr) == 1000
        assert len(bids_arr) == 1000

    @pytest.mark.slow
    def test_snapshot_exceeding_capacity_truncates(self):
        """Verify truncation to capacity."""
        book = _mk_book(num_levels=64)

        # Create 20 levels (exceeds capacity of 10)
        ask_prices = [100.0 + i * 0.01 for i in range(20)]
        bid_prices = [99.99 - i * 0.01 for i in range(20)]
        sizes = [1.0] * 20
        norders = [1] * 20

        asks = OrderbookLevels.from_list_with_ticks_and_lots(
            ask_prices, sizes, norders, TICK_SIZE, LOT_SIZE
        )
        bids = OrderbookLevels.from_list_with_ticks_and_lots(
            bid_prices, sizes, norders, TICK_SIZE, LOT_SIZE
        )

        book.consume_snapshot(asks, bids)

        asks_arr = book.get_asks_numpy()
        bids_arr = book.get_bids_numpy()

        # Should truncate to 16 levels
        assert len(asks_arr) <= 64
        assert len(bids_arr) <= 64

    def test_delta_insertion_at_full_capacity(self):
        """Worst level eviction when at capacity."""
        book = _mk_book(num_levels=64)

        # Fill to capacity with 64 ask levels
        asks, _ = _make_levels(
            [100.0 + i * 0.01 for i in range(64)], [1.0] * 64, with_precision=True
        )
        bids, _ = _make_levels([99.90], [1.0], with_precision=True)
        book.consume_snapshot(asks, bids)

        # Insert new best ask (should evict worst ask at 100.63)
        new_ask = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.98], [2.0], [1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(new_ask, _empty_bid_levels())

        asks_arr = book.get_asks_numpy()
        assert len(asks_arr) == 64  # Still at capacity
        assert asks_arr["price"][0] == pytest.approx(99.98)  # New best ask

    def test_roll_right_at_max_capacity(self):
        """BBO insertion evicts last level."""
        book = _mk_book(num_levels=64)

        # Fill to capacity with 64 levels
        asks, _ = _make_levels(
            [100.0 + i * 0.01 for i in range(64)],
            [float(i + 1) for i in range(64)],
            with_precision=True,
        )
        bids, _ = _make_levels([99.90], [1.0], with_precision=True)
        book.consume_snapshot(asks, bids)

        # Insert better ask at front (should evict last level)
        better_ask = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.95], [5.0], [1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(better_ask, _empty_bid_levels())

        asks_arr = book.get_asks_numpy()
        assert len(asks_arr) == 64
        assert asks_arr["price"][0] == pytest.approx(99.95)
        # Last level should be evicted

    @pytest.mark.slow
    def test_rapid_insertions_and_deletions_at_capacity(self):
        """1000 iteration stress test at capacity."""
        book = _mk_book(num_levels=64)

        # Initial snapshot
        asks, _ = _make_levels(
            [100.0 + i * 0.01 for i in range(10)], [1.0] * 10, with_precision=True
        )
        bids, _ = _make_levels(
            [99.99 - i * 0.01 for i in range(10)], [1.0] * 10, with_precision=True
        )
        book.consume_snapshot(asks, bids)

        # Rapid updates
        for i in range(1000):
            price = 100.0 + (i % 20) * 0.01
            size = 1.0 if i % 2 == 0 else 0.0  # Alternate insert/delete

            delta = OrderbookLevels.from_list_with_ticks_and_lots(
                [price], [size], [1 if size > 0 else 0], TICK_SIZE, LOT_SIZE
            )
            book.consume_deltas(delta, _empty_bid_levels())

        # Should still be valid
        asks_arr = book.get_asks_numpy()
        assert len(asks_arr) <= 64


@pytest.mark.boundary
class TestPathologicalDataBoundaries:
    """Test pathological data scenarios."""

    def test_huge_spread_between_bbo(self, pathological_data):
        """bid=1.00, ask=10000.00."""
        book = _mk_book(num_levels=64)

        bid_price, ask_price = pathological_data["huge_spread"]

        asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [ask_price], [1.0], [1], TICK_SIZE, LOT_SIZE
        )
        bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [bid_price], [1.0], [1], TICK_SIZE, LOT_SIZE
        )

        book.consume_snapshot(asks, bids)

        spread = book.get_bbo_spread()
        assert spread == pytest.approx(ask_price - bid_price)

    def test_zero_tick_spread(self, pathological_data):
        """bid=ask=100.00."""
        book = _mk_book(num_levels=64)

        bid_price, ask_price = pathological_data["zero_spread"]

        asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [ask_price], [1.0], [1], TICK_SIZE, LOT_SIZE
        )
        bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [bid_price], [1.0], [1], TICK_SIZE, LOT_SIZE
        )

        book.consume_snapshot(asks, bids)

        spread = book.get_bbo_spread()
        assert spread == pytest.approx(0.0)

    def test_negative_spread_via_snapshot(self, pathological_data):
        """Crossed orderbook via snapshot."""
        book = _mk_book(num_levels=64)

        bid_price, ask_price = pathological_data["negative_spread"]

        # Crossed: bid > ask
        asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [ask_price], [1.0], [1], TICK_SIZE, LOT_SIZE
        )
        bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [bid_price], [1.0], [1], TICK_SIZE, LOT_SIZE
        )

        book.consume_snapshot(asks, bids)

        # After cross removal, orderbook should be empty or have non-crossed levels
        try:
            spread = book.get_bbo_spread()
            assert spread >= 0.0  # Should not be negative
        except RuntimeError:
            # Empty after cross removal is also valid
            pass

    def test_duplicate_price_levels_in_snapshot(self):
        """Deduplication verification."""
        book = _mk_book(num_levels=64)

        # Duplicate ask prices
        asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.0, 100.0, 100.01], [1.0, 2.0, 1.0], [1, 1, 1], TICK_SIZE, LOT_SIZE
        )
        bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.99], [1.0], [1], TICK_SIZE, LOT_SIZE
        )

        book.consume_snapshot(asks, bids)

        asks_arr = book.get_asks_numpy()
        # Should handle duplicates (last one wins or aggregate)
        assert len(asks_arr) > 0

    def test_all_zero_sizes_in_snapshot(self):
        """Snapshot with all deletion markers retains zero-size levels."""
        book = _mk_book(num_levels=64)

        # All zero sizes
        asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.0, 100.01], [0.0, 0.0], [0, 0], TICK_SIZE, LOT_SIZE
        )
        bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.99], [0.0], [0], TICK_SIZE, LOT_SIZE
        )

        book.consume_snapshot(asks, bids)

        # Orderbook retains zero-size levels from snapshots
        bids_arr = book.get_bids_numpy()
        asks_arr = book.get_asks_numpy()
        assert len(bids_arr) == 1
        assert len(asks_arr) == 2
        # All sizes should be zero
        assert all(bids_arr["size"] == 0.0)
        assert all(asks_arr["size"] == 0.0)

    def test_extreme_price_values(self, pathological_data):
        """prices near float64 limits (1e100, 1e-100)."""
        book = _mk_book(num_levels=64)

        extreme_high = pathological_data["extreme_price_high"]
        extreme_low = pathological_data["extreme_price_low"]

        asks = OrderbookLevels.from_list([extreme_high], [1.0], [1])
        bids = OrderbookLevels.from_list([extreme_low], [1.0], [1])

        book.consume_snapshot(asks, bids)

        asks_arr = book.get_asks_numpy()
        bids_arr = book.get_bids_numpy()

        assert asks_arr["price"][0] == pytest.approx(extreme_high)
        assert bids_arr["price"][0] == pytest.approx(extreme_low)

    def test_extreme_size_values(self, pathological_data):
        """size=1e15."""
        book = _mk_book(num_levels=64)

        extreme_size = pathological_data["extreme_size"]

        asks = OrderbookLevels.from_list([100.0], [extreme_size], [1])
        bids = OrderbookLevels.from_list([99.99], [extreme_size], [1])

        book.consume_snapshot(asks, bids)

        asks_arr = book.get_asks_numpy()
        assert asks_arr["size"][0] == pytest.approx(extreme_size)

    def test_extremely_small_tick_size(self):
        """tick_size=1e-10."""
        book = AdvancedOrderbook(
            tick_size=1e-10,
            lot_size=LOT_SIZE,
            num_levels=64,
            delta_sortedness=PyOrderbookSortedness.UNKNOWN,
            snapshot_sortedness=PyOrderbookSortedness.UNKNOWN,
        )

        asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.0], [1.0], [1], 1e-10, LOT_SIZE
        )
        bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.99], [1.0], [1], 1e-10, LOT_SIZE
        )

        book.consume_snapshot(asks, bids)

        mid_price = book.get_mid_price()
        assert mid_price > 0

    def test_extremely_large_tick_size(self):
        """tick_size=1000.0."""
        book = AdvancedOrderbook(
            tick_size=1000.0,
            lot_size=LOT_SIZE,
            num_levels=64,
            delta_sortedness=PyOrderbookSortedness.UNKNOWN,
            snapshot_sortedness=PyOrderbookSortedness.UNKNOWN,
        )

        asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [100000.0], [1.0], [1], 1000.0, LOT_SIZE
        )
        bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [99000.0], [1.0], [1], 1000.0, LOT_SIZE
        )

        book.consume_snapshot(asks, bids)

        spread = book.get_bbo_spread()
        assert spread > 0

    def test_mixed_zero_and_nonzero_updates(self):
        """Mixed delta types."""
        book = _mk_book(num_levels=64)

        # Initialize
        asks, _ = _make_levels(
            [100.0, 100.01, 100.02], [1.0, 1.0, 1.0], with_precision=True
        )
        bids, _ = _make_levels([99.90], [1.0], with_precision=True)
        book.consume_snapshot(asks, bids)

        # Mixed updates: delete 100.0, update 100.01, insert 100.03
        mixed_asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.0, 100.01, 100.03],
            [0.0, 2.0, 1.5],  # delete, update, insert
            [0, 1, 1],
            TICK_SIZE,
            LOT_SIZE,
        )

        book.consume_deltas(mixed_asks, _empty_bid_levels())

        asks_arr = book.get_asks_numpy()
        assert len(asks_arr) >= 2

    def test_out_of_order_snapshot_with_unknown_sortedness(self):
        """Auto-sort verification."""
        book = _mk_book(num_levels=64)

        # Out of order asks
        asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.02, 100.0, 100.01],  # Wrong order
            [1.0, 1.0, 1.0],
            [1, 1, 1],
            TICK_SIZE,
            LOT_SIZE,
        )
        bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.97, 99.99, 99.98],  # Wrong order
            [1.0, 1.0, 1.0],
            [1, 1, 1],
            TICK_SIZE,
            LOT_SIZE,
        )

        book.consume_snapshot(asks, bids)

        asks_arr = book.get_asks_numpy()
        bids_arr = book.get_bids_numpy()

        # Should be sorted after consumption
        assert asks_arr["price"][0] <= asks_arr["price"][-1]  # Ascending
        assert bids_arr["price"][0] >= bids_arr["price"][-1]  # Descending

    def test_sequential_crosses_via_bbo_updates(self):
        """Multiple cross removals."""
        book = _mk_book(num_levels=64)

        # Initialize
        asks, _ = _make_levels(
            [100.0, 100.01, 100.02], [1.0, 1.0, 1.0], with_precision=True
        )
        bids, _ = _make_levels(
            [99.99, 99.98, 99.97], [1.0, 1.0, 1.0], with_precision=True
        )
        book.consume_snapshot(asks, bids)

        # First cross: ask at 99.98 (crosses with top 2 bids)
        cross1 = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.98], [1.0], [1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(cross1, _empty_bid_levels())

        # Second cross: ask at 99.96 (should cross with remaining bids)
        cross2 = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.96], [1.0], [1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(cross2, _empty_bid_levels())

        # Should handle sequential crosses
        try:
            book.get_bids_numpy()
        except RuntimeError:
            # Empty after crosses is valid
            pass


@pytest.mark.boundary
class TestPrecisionAndRoundingBoundaries:
    """Test precision and rounding edge cases."""

    def test_tick_rounding_near_boundaries(self):
        """price=100.005 → floor to 10000 ticks."""
        # With TICK_SIZE=0.01, price=100.005 should round to 100.00 ticks (10000 ticks)
        level = OrderbookLevel.with_ticks_and_lots(
            100.005, 1.0, TICK_SIZE, LOT_SIZE, norders=1
        )

        expected_ticks = int(100.005 / TICK_SIZE)
        assert level.ticks == expected_ticks

    def test_lot_rounding_near_boundaries(self):
        """size=1.0005 rounding."""
        level = OrderbookLevel.with_ticks_and_lots(
            100.0, 1.0005, TICK_SIZE, LOT_SIZE, norders=1
        )

        expected_lots = int(1.0005 / LOT_SIZE)
        assert level.lots == expected_lots

    def test_mid_price_with_odd_tick_sum(self):
        """Integer division verification."""
        book = _mk_book(num_levels=64)

        # Prices that result in odd tick sum
        asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.01], [1.0], [1], TICK_SIZE, LOT_SIZE
        )
        bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.00], [1.0], [1], TICK_SIZE, LOT_SIZE
        )

        book.consume_snapshot(asks, bids)

        mid_price = book.get_mid_price()
        expected_mid = (100.01 + 100.00) / 2.0
        assert mid_price == pytest.approx(expected_mid, abs=TICK_SIZE)

    def test_spread_precision_consistency(self):
        """Tick arithmetic vs price arithmetic."""
        book = _mk_book(num_levels=64)

        asks = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.10], [1.0], [1], TICK_SIZE, LOT_SIZE
        )
        bids = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.00], [1.0], [1], TICK_SIZE, LOT_SIZE
        )

        book.consume_snapshot(asks, bids)

        spread = book.get_bbo_spread()
        expected_spread = 100.10 - 100.00
        assert spread == pytest.approx(expected_spread, abs=1e-9)


@pytest.mark.boundary
class TestCapacityOverflowProtection:
    """Test overflow protection guards."""

    def test_increment_count_at_max_capacity_noop(self):
        """Guard at ladder.pyx:125-127."""
        book = _mk_book(num_levels=64)

        # Fill to capacity with 16 levels
        asks, _ = _make_levels(
            [100.0 + i * 0.01 for i in range(16)], [1.0] * 16, with_precision=True
        )
        bids, _ = _make_levels([99.90], [1.0], with_precision=True)
        book.consume_snapshot(asks, bids)

        # Try to add beyond capacity (worst price, should be ignored)
        extra_ask = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.20], [1.0], [1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(extra_ask, _empty_bid_levels())

        asks_arr = book.get_asks_numpy()
        # Should not exceed capacity
        assert len(asks_arr) <= 64

    def test_decrement_count_at_zero_noop(self):
        """Guard at ladder.pyx:130-133."""
        book = _mk_book(num_levels=64)

        # Initialize with one level
        asks, _ = _make_levels([100.0], [1.0], with_precision=True)
        bids, _ = _make_levels([99.99], [1.0], with_precision=True)
        book.consume_snapshot(asks, bids)

        # Delete the level
        delete_ask = OrderbookLevels.from_list_with_ticks_and_lots(
            [100.0], [0.0], [0], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(delete_ask, _empty_bid_levels())

        # Try to delete again (should be no-op)
        book.consume_deltas(delete_ask, _empty_bid_levels())

        # Should not crash, and asks should be empty
        asks_arr = book.get_asks_numpy()
        assert len(asks_arr) == 0

    def test_roll_right_never_writes_past_capacity(self):
        """Verify roll operations respect capacity."""
        book = _mk_book(num_levels=64)

        # Fill to capacity with 16 levels
        asks, _ = _make_levels(
            [100.0 + i * 0.01 for i in range(16)],
            [1.0] * 16,
            with_precision=True,
        )
        bids, _ = _make_levels([99.90], [1.0], with_precision=True)
        book.consume_snapshot(asks, bids)

        # Insert at front (triggers roll_right, should evict last)
        new_ask = OrderbookLevels.from_list_with_ticks_and_lots(
            [99.95], [2.0], [1], TICK_SIZE, LOT_SIZE
        )
        book.consume_deltas(new_ask, _empty_bid_levels())

        asks_arr = book.get_asks_numpy()
        assert len(asks_arr) <= 64
        assert asks_arr["price"][0] == pytest.approx(99.95)



from libc.stdint cimport int64_t as i64

cdef struct SubEventTokenState:
    i64 allocated_tokens
    i64 used_tokens
    i64 prev_refill_time_ms
    i64 next_refill_time_ms

cdef struct EventTokenState:
    i64 allocated_tokens
    i64 used_tokens
    i64 prev_refill_time_ms
    i64 next_refill_time_ms
    i64 burst_attempts_used
    i64 num_sub_event_states
    SubEventTokenState* sub_event_states




"""Tests for rate limiter utilities."""

from __future__ import annotations

import time

import pytest

from mm_toolbox.rate_limiter import (
    RateLimitBurstConfig,
    RateLimiter,
    RateLimiterConfig,
    RateLimitState,
    RateLimitStateConfig,
    SubBucketStrategy,
)


def _avoid_second_boundary(buffer_s: float = 0.05) -> None:
    """Sleep briefly to avoid crossing a second boundary mid-test."""
    if time.time() % 1.0 > 1.0 - buffer_s:
        time.sleep(buffer_s * 2.0)


def make_limiter(
    capacity: int,
    window_s: int,
    *,
    warn: float = 0.75,
    block: float = 0.95,
    state_enabled: bool = True,
    burst_enabled: bool = False,
    max_tokens: int = 0,
    max_attempts: int = 0,
    sub_bucket_strategy: SubBucketStrategy = SubBucketStrategy.PER_SECOND,
) -> RateLimiter:
    """Create a limiter with explicit configuration controls."""
    state_cfg = RateLimitStateConfig(
        is_enabled=state_enabled,
        warning_threshold=warn,
        block_threshold=block,
    )
    burst_cfg = RateLimitBurstConfig(
        is_enabled=burst_enabled,
        max_tokens=max_tokens,
        max_burst_attempts=max_attempts,
    )
    cfg = RateLimiterConfig(
        capacity=capacity,
        window_s=window_s,
        state_config=state_cfg,
        burst_config=burst_cfg,
        sub_bucket_strategy=sub_bucket_strategy,
    )
    return RateLimiter(cfg)


class TestRateLimitStateConfig:
    """Test RateLimitStateConfig validation and defaults."""

    def test_default_thresholds(self) -> None:
        """Default thresholds are enabled and ordered."""
        cfg = RateLimitStateConfig.default()
        assert cfg.is_enabled is True
        assert cfg.warning_threshold == 0.75
        assert cfg.block_threshold == 0.95

    @pytest.mark.parametrize(
        "warning_threshold, block_threshold",
        [
            (0.0, 0.5),
            (1.0, 0.5),
            (0.5, 0.0),
            (0.5, 1.0),
            (0.8, 0.7),
        ],
    )
    def test_invalid_thresholds(
        self, warning_threshold: float, block_threshold: float
    ) -> None:
        """Thresholds must be in (0, 1) and strictly increasing."""
        with pytest.raises(ValueError):
            RateLimitStateConfig(
                is_enabled=True,
                warning_threshold=warning_threshold,
                block_threshold=block_threshold,
            )


class TestRateLimitBurstConfig:
    """Test RateLimitBurstConfig validation and defaults."""

    def test_default_config(self) -> None:
        """Default burst config is disabled."""
        cfg = RateLimitBurstConfig.default()
        assert cfg.is_enabled is False
        assert cfg.max_tokens == 0
        assert cfg.max_burst_attempts == 0

    @pytest.mark.parametrize("max_tokens, max_attempts", [(0, 1), (1, 0), (-1, 2)])
    def test_invalid_enabled_settings(self, max_tokens: int, max_attempts: int) -> None:
        """Enabled burst config requires positive limits."""
        with pytest.raises(ValueError):
            RateLimitBurstConfig(
                is_enabled=True,
                max_tokens=max_tokens,
                max_burst_attempts=max_attempts,
            )

    def test_valid_enabled_settings(self) -> None:
        """Enabled burst config accepts positive limits."""
        cfg = RateLimitBurstConfig(
            is_enabled=True,
            max_tokens=2,
            max_burst_attempts=1,
        )
        assert cfg.is_enabled is True
        assert cfg.max_tokens == 2
        assert cfg.max_burst_attempts == 1


class TestRateLimiterConfig:
    """Test RateLimiterConfig validation and defaults."""

    def test_default_config(self) -> None:
        """Default config wires state and burst policies."""
        cfg = RateLimiterConfig.default(capacity=5, window_s=2)
        assert cfg.capacity == 5
        assert cfg.window_s == 2
        assert cfg.state_config.is_enabled is True
        assert cfg.burst_config.is_enabled is False
        assert cfg.sub_bucket_strategy is SubBucketStrategy.PER_SECOND

    @pytest.mark.parametrize("capacity, window_s", [(0, 1), (-1, 1), (1, 0), (1, -2)])
    def test_invalid_capacity_or_window(self, capacity: int, window_s: int) -> None:
        """Capacity and window duration must be positive."""
        with pytest.raises(ValueError):
            RateLimiterConfig.default(capacity=capacity, window_s=window_s)


class TestRateLimiterBasicOperations:
    """Test core consume and accounting behavior."""

    def test_basic_consumption_and_usage(self) -> None:
        """Consuming tokens updates remaining count and usage."""
        rl = make_limiter(
            capacity=4,
            window_s=1,
            state_enabled=False,
            sub_bucket_strategy=SubBucketStrategy.DISABLED,
        )

        first = rl.try_consume()
        assert first.allowed is True
        assert first.state == RateLimitState.NORMAL
        assert first.remaining == 3
        assert first.usage == pytest.approx(0.25)

        no_change = rl.try_consume_multiple(0)
        assert no_change.allowed is True
        assert no_change.state == RateLimitState.NORMAL
        assert no_change.remaining == 3
        assert no_change.usage == pytest.approx(0.25)

        negative = rl.try_consume_multiple(-2)
        assert negative.allowed is True
        assert negative.remaining == 3
        assert negative.usage == pytest.approx(0.25)

        assert rl.tokens_remaining() == 3
        assert rl.usage() == pytest.approx(0.25)

    def test_force_consumption_over_capacity(self) -> None:
        """Force mode bypasses checks and marks OVERRIDE."""
        rl = make_limiter(
            capacity=2,
            window_s=1,
            state_enabled=False,
            sub_bucket_strategy=SubBucketStrategy.DISABLED,
        )

        forced = rl.try_consume_multiple(3, force=True)
        assert forced.allowed is True
        assert forced.state == RateLimitState.OVERRIDE
        assert forced.remaining == -1
        assert forced.usage == pytest.approx(1.5)
        assert rl.tokens_remaining() == -1


class TestRateLimiterThresholds:
    """Test warning and blocking thresholds."""

    def test_warning_and_block_transitions(self) -> None:
        """Warning triggers at > warn and block triggers at > block."""
        rl = make_limiter(
            capacity=4,
            window_s=1,
            warn=0.5,
            block=0.75,
            state_enabled=True,
            sub_bucket_strategy=SubBucketStrategy.DISABLED,
        )

        normal = rl.try_consume_multiple(2)
        assert normal.allowed is True
        assert normal.state == RateLimitState.NORMAL
        assert normal.usage == pytest.approx(0.5)

        warning = rl.try_consume()
        assert warning.allowed is True
        assert warning.state == RateLimitState.WARNING
        assert warning.usage == pytest.approx(0.75)

        blocked = rl.try_consume()
        assert blocked.allowed is False
        assert blocked.state == RateLimitState.BLOCKED
        assert blocked.remaining == 1
        assert blocked.usage == pytest.approx(0.75)
        assert rl.tokens_remaining() == 1


class TestRateLimiterSubBuckets:
    """Test per-second sub-bucket behavior."""

    def test_sub_bucket_limits_same_second(self) -> None:
        """Sub-buckets cap per-second usage even when capacity remains."""
        _avoid_second_boundary()
        rl = make_limiter(
            capacity=4,
            window_s=2,
            state_enabled=False,
            sub_bucket_strategy=SubBucketStrategy.PER_SECOND,
        )

        allowed = rl.try_consume_multiple(2)
        assert allowed.allowed is True
        assert allowed.remaining == 2

        blocked = rl.try_consume()
        assert blocked.allowed is False
        assert blocked.state == RateLimitState.BLOCKED
        assert blocked.remaining == 2

        time.sleep(1.05)
        next_bucket = rl.try_consume_multiple(2)
        assert next_bucket.allowed is True
        assert next_bucket.remaining == 0


class TestRateLimiterBurst:
    """Test burst allowance behavior."""

    def test_burst_allows_limited_overage(self) -> None:
        """Burst allows one extra attempt within configured limits."""
        rl = make_limiter(
            capacity=2,
            window_s=2,
            state_enabled=False,
            burst_enabled=True,
            max_tokens=2,
            max_attempts=1,
            sub_bucket_strategy=SubBucketStrategy.PER_SECOND,
        )

        burst_ok = rl.try_consume_multiple(2)
        assert burst_ok.allowed is True
        assert burst_ok.state == RateLimitState.NORMAL
        assert burst_ok.remaining == 0

        burst_exhausted = rl.try_consume()
        assert burst_exhausted.allowed is False
        assert burst_exhausted.state == RateLimitState.WARNING
        assert burst_exhausted.remaining == 0


class TestRateLimiterFactories:
    """Test RateLimiter factory constructors."""

    def test_factory_constructors(self) -> None:
        """Factory methods return functional limiters with expected capacity."""
        per_second = RateLimiter.per_second(3)
        per_minute = RateLimiter.per_minute(4)
        per_window = RateLimiter.per_window(5, 2)

        assert per_second.tokens_remaining() == 3
        assert per_minute.tokens_remaining() == 4
        assert per_window.tokens_remaining() == 5


class TestRateLimiterIntegration:
    """Test refill behavior in a realistic loop."""

    def test_refill_resets_after_window(self, wait_for) -> None:
        """Tokens are restored after the window elapses."""
        rl = make_limiter(
            capacity=2,
            window_s=1,
            state_enabled=False,
            sub_bucket_strategy=SubBucketStrategy.DISABLED,
        )

        consumed = rl.try_consume_multiple(2)
        assert consumed.allowed is True
        assert rl.tokens_remaining() == 0

        assert wait_for(lambda: rl.tokens_remaining() == 2, timeout_s=2.0)



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True

"""SPSC shared-memory bytes ring buffer.

Design
-------
- Single-producer/single-consumer over a file-backed mmap.
- Messages are length-prefixed (u64, little-endian) allowing arbitrary byte payloads.
- Producer advances write_pos with release semantics; consumer reads write_pos with acquire semantics.
- Consumer advances read_pos with release semantics; producer reads read_pos with acquire semantics.
- Capacity is rounded to the next power of two; indices use a mask to wrap.

Notes
-----
- Inserts are non-blocking and overwrite the oldest messages as needed.
- Consumer operations may block while waiting for new data.
- This is intentionally ONLY SPSC (single-producer/single-consumer) rings. A wrapper for SPMC will be added in the future,
  wrapping multiple SPSC rings within the same API.
"""

import os
import time
from libc.stdint cimport uint64_t as u64
from libc.string cimport memcpy
from libc.stddef cimport size_t
from libc.errno cimport errno

cdef extern from "sys/mman.h":
    void* mmap(void* addr, size_t length, int prot, int flags, int fd, long offset)
    int munmap(void* addr, size_t length)
    int PROT_READ
    int PROT_WRITE
    int MAP_SHARED

cdef extern from "fcntl.h":
    int open(const char* path, int oflag, ...)
    int O_RDONLY
    int O_RDWR
    int O_CREAT
    int O_EXCL

cdef extern from "unistd.h":
    int ftruncate(int fd, long length)
    int close(int fd)

# Use compiler builtins for atomics to avoid requiring stdatomic flags
cdef extern from *:
    unsigned long long __atomic_load_n(unsigned long long* ptr, int memorder)
    void __atomic_store_n(unsigned long long* ptr, unsigned long long val, int memorder)

cdef int _ATOMIC_RELAXED = 0
cdef int _ATOMIC_ACQUIRE = 2
cdef int _ATOMIC_RELEASE = 3

from .shm cimport ShmHeader, _SharedBytesRing

cdef inline u64 _atomic_load_acquire(u64* p) nogil:
    """Load value from pointer with acquire semantics."""
    return <u64>__atomic_load_n(<unsigned long long*>p, _ATOMIC_ACQUIRE)

cdef inline void _atomic_store_release(u64* p, u64 v) nogil:
    """Store value to pointer with release semantics."""
    __atomic_store_n(<unsigned long long*>p, <unsigned long long>v, _ATOMIC_RELEASE)

cdef inline size_t _align_up(size_t x, size_t a) nogil:
    """Align value up to the next multiple of alignment."""
    return (x + (a - 1)) & ~(a - 1)

cdef inline u64 _pow2_at_least(u64 v) nogil:
    """Return the smallest power of two greater than or equal to value."""
    cdef u64 p = 1
    if v <= 1:
        return 1
    while p < v:
        p <<= 1
    return p

cdef inline void _write_u64_le(unsigned char* base, u64 idx, u64 mask, u64 val) nogil:
    """Write 64-bit value in little-endian format with wrap-around."""
    cdef u64 pos = idx & mask
    cdef u64 capacity = mask + 1
    if pos + 8 <= capacity:
        # Fast path: no wrap-around, use direct write if aligned
        if pos & 7 == 0:
            # Aligned: direct 64-bit write
            (<u64*>(base + pos))[0] = val
        else:
            # Unaligned: byte-by-byte
            base[pos + 0] = <unsigned char>(val & 0xFF)
            base[pos + 1] = <unsigned char>((val >> 8) & 0xFF)
            base[pos + 2] = <unsigned char>((val >> 16) & 0xFF)
            base[pos + 3] = <unsigned char>((val >> 24) & 0xFF)
            base[pos + 4] = <unsigned char>((val >> 32) & 0xFF)
            base[pos + 5] = <unsigned char>((val >> 40) & 0xFF)
            base[pos + 6] = <unsigned char>((val >> 48) & 0xFF)
            base[pos + 7] = <unsigned char>((val >> 56) & 0xFF)
    else:
        # Slow path: wrap-around
        base[(idx + 0) & mask] = <unsigned char>(val & 0xFF)
        base[(idx + 1) & mask] = <unsigned char>((val >> 8) & 0xFF)
        base[(idx + 2) & mask] = <unsigned char>((val >> 16) & 0xFF)
        base[(idx + 3) & mask] = <unsigned char>((val >> 24) & 0xFF)
        base[(idx + 4) & mask] = <unsigned char>((val >> 32) & 0xFF)
        base[(idx + 5) & mask] = <unsigned char>((val >> 40) & 0xFF)
        base[(idx + 6) & mask] = <unsigned char>((val >> 48) & 0xFF)
        base[(idx + 7) & mask] = <unsigned char>((val >> 56) & 0xFF)

cdef inline u64 _read_u64_le(const unsigned char* base, u64 idx, u64 mask) nogil:
    """Read 64-bit value in little-endian format with wrap-around."""
    cdef u64 pos = idx & mask
    cdef u64 capacity = mask + 1
    if pos + 8 <= capacity:
        # Fast path: no wrap-around
        if pos & 7 == 0:
            # Aligned: direct 64-bit read
            return (<u64*>(base + pos))[0]
        else:
            # Unaligned: byte-by-byte
            return (
                (<u64>base[pos + 0])
                | (<u64>base[pos + 1] << 8)
                | (<u64>base[pos + 2] << 16)
                | (<u64>base[pos + 3] << 24)
                | (<u64>base[pos + 4] << 32)
                | (<u64>base[pos + 5] << 40)
                | (<u64>base[pos + 6] << 48)
                | (<u64>base[pos + 7] << 56)
            )
    else:
        # Slow path: wrap-around
        return (
            (<u64>base[(idx + 0) & mask])
            | (<u64>base[(idx + 1) & mask] << 8)
            | (<u64>base[(idx + 2) & mask] << 16)
            | (<u64>base[(idx + 3) & mask] << 24)
            | (<u64>base[(idx + 4) & mask] << 32)
            | (<u64>base[(idx + 5) & mask] << 40)
            | (<u64>base[(idx + 6) & mask] << 48)
            | (<u64>base[(idx + 7) & mask] << 56)
        )

cdef inline void _copy_in(unsigned char* dst, const unsigned char* src, size_t n) nogil:
    """Copy bytes from source to destination."""
    memcpy(dst, src, n)

cdef inline void _copy_into_ring(unsigned char* ring, u64 start, u64 mask, const unsigned char* src, size_t n, u64 capacity) nogil:
    """Copy bytes into ring buffer with wrap-around."""
    if n == 0:
        return
    cdef u64 idx = start & mask
    cdef u64 endspace = capacity - idx
    if n <= endspace:
        # Fast path: no wrap-around
        memcpy(ring + idx, src, n)
    else:
        # Slow path: wrap-around
        memcpy(ring + idx, src, <size_t>endspace)
        memcpy(ring, src + <size_t>endspace, n - <size_t>endspace)

cdef inline void _copy_from_ring(unsigned char* dst, const unsigned char* ring, u64 start, u64 mask, size_t n, u64 capacity) nogil:
    """Copy bytes from ring buffer with wrap-around."""
    if n == 0:
        return
    cdef u64 idx = start & mask
    cdef u64 endspace = capacity - idx
    if n <= endspace:
        # Fast path: no wrap-around
        memcpy(dst, ring + idx, n)
    else:
        # Slow path: wrap-around
        memcpy(dst, ring + idx, <size_t>endspace)
        memcpy(dst + <size_t>endspace, ring, n - <size_t>endspace)


cdef u64 _MAGIC = 0x53484252  # 'SHBR' (Shared Bytes Ring)
cdef u64 _VERSION = 1
cdef size_t _HEADER_ALIGN = 64
cdef size_t _HEADER_SIZE = _align_up(sizeof(ShmHeader), _HEADER_ALIGN)


cdef class _SharedBytesRing:
    """Common mapping and lifecycle for shared ring buffers.

    Attributes
    ----------
    _hdr
        Pointer to shared header containing cursors and metadata.
    _data
        Pointer to ring data region (capacity bytes).
    _capacity
        Ring capacity in bytes, power-of-two.
    _mask
        Capacity - 1 for fast wrapping.
    _spin_wait
        Spin iterations before yielding inside blocking ops.
    """
    def __cinit__(self) -> None:
        """Initialize the shared bytes ring buffer."""
        if os.name != "posix":
            raise OSError("Shared memory ringbuffer is only supported on POSIX platforms")
        self._hdr = NULL
        self._data = NULL
        self._map_len = 0
        self._fd = -1
        self._owner = False
        self._unlink_on_close = False
        self._capacity = 0
        self._mask = 0
        self._cached_read = 0
        self._cached_write = 0
        self._spin_wait = 1024
        self._path_py = None
        self._path = NULL

    cdef void _map_create(self, bytes path_b, u64 capacity_bytes, bint unlink_on_close, int spin_wait):
        """Create and initialize a new shared ring buffer."""
        cdef:
            int fd
            size_t cap = <size_t>_pow2_at_least(capacity_bytes if capacity_bytes > 0 else 1)
            size_t total_len = _HEADER_SIZE + cap
            void* base
        fd = open(path_b, O_CREAT | O_RDWR, 0o600)
        if fd < 0:
            raise OSError(errno, "open failed for shared ring")
        if ftruncate(fd, <long>total_len) != 0:
            close(fd)
            raise OSError(errno, "ftruncate failed for shared ring")
        base = mmap(NULL, total_len, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0)
        if <long>base == -1:
            close(fd)
            # Best-effort cleanup: try to unlink file if it was just created
            try:
                import os
                os.unlink(path_b)
            except Exception:
                pass
            raise OSError(errno, "mmap failed for shared ring")

        self._fd = fd
        self._map_len = total_len
        self._hdr = <ShmHeader*>base
        self._data = <unsigned char*>base + _HEADER_SIZE
        self._owner = True
        self._unlink_on_close = unlink_on_close
        self._spin_wait = spin_wait if spin_wait > 0 else 1024
        self._capacity = <u64>cap
        self._mask = <u64>(cap - 1)
        self._cached_read = 0
        self._cached_write = 0

        # Initialize header
        self._hdr.magic = _MAGIC
        self._hdr.version = _VERSION
        self._hdr.capacity = <u64>cap
        self._hdr.mask = <u64>(cap - 1)
        self._hdr.read_pos = 0
        self._hdr.write_pos = 0

    cdef void _map_attach(self, bytes path_b, int spin_wait):
        """Attach to an existing shared ring buffer."""
        cdef:
            int fd
            void* base
            size_t total_len
            u64 capacity
            u64 mask
            object backing_len
        fd = open(path_b, O_RDWR, 0o600)
        if fd < 0:
            raise OSError(errno, "open failed for shared ring")
        try:
            backing_len = os.fstat(fd).st_size
        except Exception:
            close(fd)
            raise
        if backing_len < _HEADER_SIZE:
            close(fd)
            raise RuntimeError("Shared ring backing file too small for header (legacy)")
        # Map header only to read capacity
        base = mmap(NULL, _HEADER_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0)
        if <long>base == -1:
            close(fd)
            raise OSError(errno, "mmap header failed for shared ring")
        self._hdr = <ShmHeader*>base
        if self._hdr.magic != _MAGIC or self._hdr.version != _VERSION:
            munmap(base, _HEADER_SIZE)
            close(fd)
            raise RuntimeError("Shared ring header mismatch (legacy format)")
        capacity = self._hdr.capacity
        mask = self._hdr.mask
        if capacity == 0 or (capacity & (capacity - 1)) != 0 or mask != capacity - 1:
            munmap(base, _HEADER_SIZE)
            close(fd)
            raise RuntimeError("Shared ring header invalid (legacy format)")
        if capacity > <u64>((<size_t>-1) - _HEADER_SIZE):
            munmap(base, _HEADER_SIZE)
            close(fd)
            raise RuntimeError("Shared ring header invalid (legacy format)")

        total_len = _HEADER_SIZE + <size_t>capacity
        try:
            backing_len = os.fstat(fd).st_size
        except Exception:
            munmap(base, _HEADER_SIZE)
            close(fd)
            raise
        if backing_len < total_len:
            munmap(base, _HEADER_SIZE)
            close(fd)
            raise RuntimeError("Shared ring backing file too small for header capacity (legacy)")
        munmap(base, _HEADER_SIZE)

        base = mmap(NULL, total_len, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0)
        if <long>base == -1:
            close(fd)
            raise OSError(errno, "mmap failed for shared ring")

        self._fd = fd
        self._map_len = total_len
        self._hdr = <ShmHeader*>base
        self._data = <unsigned char*>base + _HEADER_SIZE
        self._owner = False
        self._unlink_on_close = False
        self._spin_wait = spin_wait if spin_wait > 0 else 1024
        self._capacity = capacity
        self._mask = mask
        self._cached_read = self._hdr.read_pos
        self._cached_write = self._hdr.write_pos

    cdef inline void _close_map(self):
        """Unmap memory and close file descriptor."""
        if self._hdr != NULL:
            munmap(<void*>self._hdr, self._map_len)
            self._hdr = NULL
            self._data = NULL
            self._map_len = 0
        if self._fd >= 0:
            close(self._fd)
            self._fd = -1

    def __dealloc__(self):
        """Deallocate resources on object destruction."""
        try:
            self.close()
        except Exception:
            pass

    cpdef void close(self):
        """Close all mappings and optionally unlink the backing file."""
        cdef object path_obj
        cdef bytes path_b
        self._close_map()
        if self._owner and self._unlink_on_close and self._path_py is not None:
            # Best-effort unlink via Python os to avoid cross-platform libc differences
            try:
                import os
                os.unlink(self._path_py)
            except Exception:
                pass
        self._owner = False

    def __enter__(self):
        """Enter context manager, returning self.
        
        Returns:
            Self for use in with statements.
        """
        return self

    def __exit__(self, exc_type, exc, tb):
        """Exit context manager, closing resources.
        
        Args:
            exc_type: Exception type if an exception occurred.
            exc: Exception instance if an exception occurred.
            tb: Traceback if an exception occurred.
        """
        self.close()


cdef class SharedBytesRingBufferProducer(_SharedBytesRing):
    """Shared-memory SPSC producer for bytes payloads.

    Parameters
    ----------
    path
        Filesystem path for the ring backing file.
    capacity_bytes
        Total ring capacity in bytes (rounded to next power of two).
    create
        If True, create/initialize the ring; otherwise attach to existing.
    unlink_on_close
        If True and owner, unlink backing file on close.
    spin_wait
        Spin iterations in blocking calls before polling again.
    """
    def __cinit__(
        self,
        path: str,
        int capacity_bytes,
        *,
        bint create=True,
        bint unlink_on_close=False,
        int spin_wait=1024,
    ) -> None:
        """Initialize the producer.
        
        Args:
            path: Filesystem path for the ring backing file.
            capacity_bytes: Total ring capacity in bytes (rounded to next power of two).
            create: If True, create/initialize the ring; otherwise attach to existing.
            unlink_on_close: If True and owner, unlink backing file on close.
            spin_wait: Spin iterations in blocking calls before polling again.
        """
        self._path_py = path
        path_b = (<str>path).encode()
        self._path = path_b
        if create:
            self._map_create(path_b, <u64>capacity_bytes, unlink_on_close, spin_wait)
        else:
            self._map_attach(path_b, spin_wait)

    cdef inline bint _reserve(self, size_t need):
        """Ensure space for required bytes by dropping oldest messages as needed."""
        cdef u64 cap = self._capacity
        cdef u64 mask = self._mask
        cdef u64 w = self._cached_write
        cdef u64 r
        cdef u64 free_bytes
        cdef u64 L
        cdef u64 dropped_pos = 0
        with nogil:
            r = _atomic_load_acquire(&self._hdr.read_pos)
        self._cached_read = r
        free_bytes = cap - (w - r)
        if free_bytes >= <u64>need:
            return True
        # Not enough space: drop oldest messages until we have enough
        dropped_pos = r
        while True:
            L = _read_u64_le(self._data, dropped_pos & mask, mask)
            # Overflow check: ensure L doesn't exceed capacity
            if L > cap or (dropped_pos + 8 + L) < dropped_pos:
                return False
            dropped_pos += 8 + L
            free_bytes = cap - (w - dropped_pos)
            if free_bytes >= <u64>need:
                # Batch update: single atomic store for all dropped messages
                with nogil:
                    _atomic_store_release(&self._hdr.read_pos, dropped_pos)
                self._cached_read = dropped_pos
                return True

    cdef inline bint _can_reserve_once(self, size_t need):
        """Check once if space for required bytes is available."""
        cdef u64 cap = self._capacity
        cdef u64 w = self._cached_write
        cdef u64 r
        with nogil:
            r = _atomic_load_acquire(&self._hdr.read_pos)
        self._cached_read = r
        return (cap - (w - r)) >= <u64>need

    cpdef bint insert(self, bytes item):
        """Insert a single item into the ring buffer."""
        cdef:
            Py_ssize_t n = len(item)
            size_t total = <size_t>(8 + n)
            u64 cap = self._capacity
            u64 mask = self._mask
            u64 w = self._cached_write
        if n < 0:
            return False
        if total > cap:
            return False
        if not self._reserve(total):
            return False
        # Write header and payload
        _write_u64_le(self._data, w & mask, mask, <u64>n)
        _copy_into_ring(self._data, w + 8, mask, <const unsigned char*>item, <size_t>n, cap)
        w += total
        with nogil:
            _atomic_store_release(&self._hdr.write_pos, w)
        self._cached_write = w
        return True

    cpdef bint insert_batch(self, list[bytes] items):
        """Insert multiple items in order using a single reserve and commit."""
        cdef:
            Py_ssize_t i, n = len(items)
            u64 cap = self._capacity
            u64 mask = self._mask
            u64 w = self._cached_write
            u64 r
            u64 free_bytes
            u64 L64
            size_t total = 0
            bytes it
        if n == 0:
            return True
        # Validate and compute total size
        for i in range(n):
            it = items[i]
            L64 = <u64>len(it)
            if <u64>(8) + L64 > cap:
                return False
            total += <size_t>(8 + L64)
        # Drop oldest messages until enough contiguous capacity is free (byte-accurate)
        with nogil:
            r = _atomic_load_acquire(&self._hdr.read_pos)
        self._cached_read = r
        while (cap - (w - r)) < <u64>total:
            L64 = _read_u64_le(self._data, r & mask, mask)
            # Overflow check: ensure L64 doesn't exceed capacity
            if L64 > cap or (r + 8 + L64) < r:
                return False
            r += 8 + L64
            with nogil:
                _atomic_store_release(&self._hdr.read_pos, r)
            self._cached_read = r
        # Write all items
        for i in range(n):
            it = items[i]
            L64 = <u64>len(it)
            _write_u64_le(self._data, w & mask, mask, L64)
            _copy_into_ring(self._data, w + 8, mask, <const unsigned char*>it, <size_t>L64, cap)
            w += 8 + L64
        with nogil:
            _atomic_store_release(&self._hdr.write_pos, w)
        self._cached_write = w
        return True

    cpdef bint insert_packed(self, list[bytes] items):
        """Insert items as one packed message."""
        cdef:
            Py_ssize_t i, n = len(items)
            bytes it
            Py_ssize_t alloc_len
            u64 total = 0
            u64 L64
            u64 capacity = self._capacity
            bytearray buf
            unsigned char* p
            size_t off = 0
            Py_ssize_t L
            u64 py_ssize_max = <u64>((<size_t>-1) >> 1)
        if n == 0:
            return True
        if capacity <= 8:
            return False
        for i in range(n):
            it = items[i]
            L64 = <u64>len(it)
            if L64 > 0xFFFFFFFF:
                return False
            if total > (<u64>-1) - <u64>4 - L64:
                return False
            total += <u64>4 + L64
        if total <= 0:
            return True
        if total > capacity - 8:
            return False
        if total > py_ssize_max:
            return False
        alloc_len = <Py_ssize_t>total
        buf = bytearray(alloc_len)
        p = <unsigned char*>buf
        for i in range(n):
            it = items[i]
            L = len(it)
            # write 4-byte little-endian length into contiguous buffer
            p[off + 0] = <unsigned char>(L & 0xFF)
            p[off + 1] = <unsigned char>((L >> 8) & 0xFF)
            p[off + 2] = <unsigned char>((L >> 16) & 0xFF)
            p[off + 3] = <unsigned char>((L >> 24) & 0xFF)
            off += 4
            memcpy(p + off, <const unsigned char*>it, <size_t>L)
            off += <size_t>L
        # Now insert single packed message
        return self.insert(bytes(buf))


cdef class SharedBytesRingBufferConsumer(_SharedBytesRing):
    """Shared-memory SPSC consumer for bytes payloads.

    Parameters
    ----------
    path
        Filesystem path of an existing ring backing file to attach to.
    spin_wait
        Spin iterations in blocking calls before polling again.
    """
    def __cinit__(self, path: str, *, int spin_wait=1024) -> None:
        """Initialize the consumer.
        
        Args:
            path: Filesystem path of an existing ring backing file to attach to.
            spin_wait: Spin iterations in blocking calls before polling again.
        """
        self._path_py = path
        path_b = (<str>path).encode()
        self._path = path_b
        self._map_attach(path_b, spin_wait)

    cdef inline bint _peek_available(self, u64* out_len, u64* out_read_pos) nogil:
        """Check if a complete message is available without committing read."""
        cdef u64 r = _atomic_load_acquire(&self._hdr.read_pos)
        cdef u64 w = _atomic_load_acquire(&self._hdr.write_pos)
        cdef u64 avail = w - r
        if avail < 8:
            return False
        cdef u64 L = _read_u64_le(self._data, r & self._mask, self._mask)
        if avail < 8 + L:
            return False
        out_len[0] = <u64>L
        out_read_pos[0] = r
        return True

    cpdef bytes consume(self):
        """Consume a single item (blocking)."""
        cdef:
            u64 L64 = 0
            u64 r = 0
            u64 r_check = 0
            u64 mask = self._mask
            u64 cap = self._capacity
            int spin_count = 0
            bint available = False
        while True:
            with nogil:
                available = self._peek_available(&L64, &r)
            if not available:
                spin_count += 1
                if spin_count < self._spin_wait:
                    continue
                # Yield to other threads after spin_wait iterations
                time.sleep(0.0001)
                spin_count = 0
                continue
            # Re-check read_pos hasn't changed (producer might have dropped messages)
            with nogil:
                r_check = _atomic_load_acquire(&self._hdr.read_pos)
            if r_check != r:
                # Producer advanced read_pos, re-check availability
                continue
            break
        cdef u64 L = L64
        cdef bytes out = bytes(<Py_ssize_t>L)
        _copy_from_ring(<unsigned char*>out, self._data, r + 8, mask, <size_t>L, cap)
        r += 8 + L
        with nogil:
            _atomic_store_release(&self._hdr.read_pos, r)
        self._cached_read = r
        return out

    cpdef object try_consume(self):
        """Try to consume a single item (non-blocking)."""
        cdef u64 L64 = 0
        cdef u64 r = 0
        cdef u64 r_check = 0
        cdef u64 mask = self._mask
        cdef u64 cap = self._capacity
        with nogil:
            if not self._peek_available(&L64, &r):
                return None
            r_check = _atomic_load_acquire(&self._hdr.read_pos)
        if r_check != r:
            return None
        cdef u64 L = L64
        cdef bytes out = bytes(<Py_ssize_t>L)
        _copy_from_ring(<unsigned char*>out, self._data, r + 8, mask, <size_t>L, cap)
        r += 8 + L
        with nogil:
            _atomic_store_release(&self._hdr.read_pos, r)
        self._cached_read = r
        return out

    cpdef list consume_all_nowait(self):
        """Drain all currently available items without blocking."""
        cdef list res = []
        cdef object it
        while True:
            it = self.try_consume()
            if it is None:
                break
            res.append(it)
        return res

    cpdef list consume_packed(self):
        """Consume a packed message and unpack into a list of items."""
        cdef bytes buf = self.consume()
        cdef memoryview mv = memoryview(buf)
        cdef Py_ssize_t n = mv.shape[0]
        cdef Py_ssize_t off = 0
        cdef list items = []
        cdef u64 L
        while off + 4 <= n:
            L = (
                (<u64>mv[off])
                | (<u64>mv[off + 1] << 8)
                | (<u64>mv[off + 2] << 16)
                | (<u64>mv[off + 3] << 24)
            )
            off += 4
            if off + L > n:
                raise ValueError("Corrupted packed message")
            items.append(bytes(mv[off : off + L]))
            off += L
        return items



from __future__ import annotations

from collections.abc import AsyncIterator, Coroutine, Iterator
from typing import Any

class BytesRingBuffer:
    def __init__(
        self,
        max_capacity: int,
        disable_async: bool = False,
        only_insert_unique: bool = False,
    ) -> None: ...
    def raw(self, copy: bool = True) -> list[bytes]: ...
    def unwrapped(self) -> list[bytes]: ...
    def overwrite_latest(self, item: bytes, increment_count: bool = False) -> None: ...
    def insert(self, item: bytes) -> None: ...
    def insert_batch(self, items: list[bytes]) -> None: ...
    def contains(self, item: bytes) -> bool: ...
    def consume(self) -> bytes: ...
    def consume_all(self) -> list[bytes]: ...
    def consume_iterable(self) -> Iterator[bytes]: ...
    def aconsume(self) -> Coroutine[Any, Any, bytes]: ...
    def aconsume_iterable(self) -> AsyncIterator[bytes]: ...
    def peekright(self) -> bytes: ...
    def peekleft(self) -> bytes: ...
    def clear(self) -> None: ...
    def is_empty(self) -> bool: ...
    def is_full(self) -> bool: ...
    def __contains__(self, item: bytes) -> bool: ...
    def __len__(self) -> int: ...
    def __getitem__(self, idx: int) -> bytes: ...

class BytesRingBufferFast:
    def __init__(
        self,
        max_capacity: int,
        disable_async: bool = False,
        only_insert_unique: bool = False,
        expected_item_size: int = 128,
        buffer_percent: float = 25.0,
    ) -> None: ...
    def raw(self, copy: bool = True) -> list[bytes]: ...
    def unwrapped(self) -> list[bytes]: ...
    def overwrite_latest(self, item: bytes, increment_count: bool = False) -> None: ...
    def insert(self, item: bytes) -> None: ...
    def insert_char(self, item: bytes, item_len: int) -> None: ...
    def insert_batch(self, items: list[bytes]) -> None: ...
    def contains(self, item: bytes) -> bool: ...
    def consume(self) -> bytes: ...
    def consume_all(self) -> list[bytes]: ...
    def consume_iterable(self) -> Iterator[bytes]: ...
    def aconsume(self) -> Coroutine[Any, Any, bytes]: ...
    def aconsume_iterable(self) -> AsyncIterator[bytes]: ...
    def peekright(self) -> bytes: ...
    def peekleft(self) -> bytes: ...
    def clear(self) -> None: ...
    def is_empty(self) -> bool: ...
    def is_full(self) -> bool: ...
    def __contains__(self, item: bytes) -> bool: ...
    def __len__(self) -> int: ...
    def __getitem__(self, idx: int) -> bytes: ...



cimport numpy as cnp

from mm_toolbox.ringbuffer.numeric cimport NumericRingBuffer

cdef class MovingAverage:
    cdef:
        int                 _window
        bint                _is_fast
        bint                _is_warm
        double              _value
        NumericRingBuffer   _values

    # def                   __init__(self, int window, bint is_fast=False)

    cpdef double            initialize(self, cnp.ndarray values)
    cpdef double            next(self, double value) 
    cpdef double            update(self, double value)
    cpdef double            get_value(self)
    cpdef cnp.ndarray       get_values(self)

    # def __len__(self)
    # def __iter__(self)
    # def __getitem__(self, int idx)
    cdef inline void        ensure_warm(self)
    cdef inline void        push_to_ringbuffer(self)
    cdef inline void        __enforce_moving_average_initialized(self)
    cdef inline void        __enforce_not_fast(self)



"""Async iteration tests for WsSingle."""

from __future__ import annotations

import asyncio

import pytest

from mm_toolbox.websocket.single import WsSingle


@pytest.mark.asyncio
class TestWsSingleAsyncIteration:
    """Validate async iteration over WsSingle messages."""

    async def test_async_for_loop_consumes_messages(
        self,
        basic_server,
        connection_config_factory,
    ) -> None:
        """Ensure async for receives messages from the server.

        Args:
            basic_server: Fixture providing a basic echo server.
            connection_config_factory: Fixture providing config factory.

        Returns:
            None: This test does not return a value.
        """
        async with basic_server:
            config = connection_config_factory(basic_server)
            collected: list[bytes] = []

            async with WsSingle(config) as ws:

                async def _collector() -> None:
                    """Collect a fixed number of messages from iteration.

                    Returns:
                        None: This helper does not return a value.
                    """
                    async for msg in ws:
                        collected.append(msg)
                        if len(collected) >= 3:
                            break

                collect_task = asyncio.create_task(_collector())
                await asyncio.sleep(0.1)
                await basic_server.send_to_all_clients(b"msg1")
                await basic_server.send_to_all_clients(b"msg2")
                await basic_server.send_to_all_clients(b"msg3")
                await asyncio.wait_for(collect_task, timeout=2.0)

            assert len(collected) == 3

    async def test_async_iteration_arrival_validation(
        self,
        basic_server,
        connection_config_factory,
    ) -> None:
        """Ensure all sent messages arrive during iteration.

        Args:
            basic_server: Fixture providing a basic echo server.
            connection_config_factory: Fixture providing config factory.

        Returns:
            None: This test does not return a value.
        """
        async with basic_server:
            config = connection_config_factory(basic_server)
            payloads = [b"a", b"b", b"c", b"d"]
            collected: list[bytes] = []

            async with WsSingle(config) as ws:

                async def _collector() -> None:
                    """Collect all expected messages from iteration.

                    Returns:
                        None: This helper does not return a value.
                    """
                    async for msg in ws:
                        collected.append(msg)
                        if len(collected) >= len(payloads):
                            break

                collect_task = asyncio.create_task(_collector())
                await asyncio.sleep(0.1)
                for payload in payloads:
                    await basic_server.send_to_all_clients(payload)
                await asyncio.wait_for(collect_task, timeout=2.0)

            assert set(collected) == set(payloads)

    async def test_async_iteration_stops_on_disconnect(
        self,
        server_send_close_frame,
        connection_config_factory,
    ) -> None:
        """Ensure iteration can be stopped after disconnect.

        Args:
            server_send_close_frame: Fixture providing close-frame server.
            connection_config_factory: Fixture providing config factory.

        Returns:
            None: This test does not return a value.
        """
        async with server_send_close_frame:
            config = connection_config_factory(server_send_close_frame)
            async with WsSingle(config) as ws:
                collect_task = asyncio.create_task(ws.__anext__())
                await asyncio.sleep(0.1)
                ws.send_data(b"trigger-close")
                await asyncio.sleep(0.2)
                if collect_task.done():
                    assert collect_task.result() == b"trigger-close"
                else:
                    collect_task.cancel()
                    with pytest.raises(asyncio.CancelledError):
                        await collect_task

    async def test_async_iteration_concurrent_with_send(
        self,
        basic_server,
        connection_config_factory,
    ) -> None:
        """Ensure iteration and send_data can run concurrently.

        Args:
            basic_server: Fixture providing a basic echo server.
            connection_config_factory: Fixture providing config factory.

        Returns:
            None: This test does not return a value.
        """
        async with basic_server:
            config = connection_config_factory(basic_server)
            expected = {b"srv1", b"srv2"}
            seen: set[bytes] = set()

            async with WsSingle(config) as ws:

                async def _collector() -> None:
                    """Collect two messages while sending concurrently.

                    Returns:
                        None: This helper does not return a value.
                    """
                    async for msg in ws:
                        if msg in expected:
                            seen.add(msg)
                        if expected.issubset(seen):
                            break

                collect_task = asyncio.create_task(_collector())
                await asyncio.sleep(0.1)
                ws.send_data(b"client-msg")
                await basic_server.send_to_all_clients(b"srv1")
                await basic_server.send_to_all_clients(b"srv2")
                await asyncio.wait_for(collect_task, timeout=2.0)

            assert seen == expected
            assert b"client-msg" in basic_server.get_received_messages()

    async def test_async_iterator_with_large_message_burst(
        self,
        basic_server,
        connection_config_factory,
    ) -> None:
        """Ensure bursts of messages are consumed correctly.

        Args:
            basic_server: Fixture providing a basic echo server.
            connection_config_factory: Fixture providing config factory.

        Returns:
            None: This test does not return a value.
        """
        async with basic_server:
            config = connection_config_factory(basic_server)
            payloads = [f"msg-{idx}".encode("utf-8") for idx in range(10)]
            collected: list[bytes] = []

            async with WsSingle(config) as ws:

                async def _collector() -> None:
                    """Collect a burst of messages from iteration.

                    Returns:
                        None: This helper does not return a value.
                    """
                    async for msg in ws:
                        collected.append(msg)
                        if len(collected) >= len(payloads):
                            break

                collect_task = asyncio.create_task(_collector())
                await asyncio.sleep(0.1)
                for payload in payloads:
                    await basic_server.send_to_all_clients(payload)
                await asyncio.wait_for(collect_task, timeout=3.0)

            assert set(collected) == set(payloads)



cimport cython
from libc.stdint cimport uint64_t as u64

from .level.level cimport OrderbookLevels, OrderbookLevel
from .ladder.ladder cimport OrderbookLadder, OrderbookLadderData
from .enum.enums cimport CyOrderbookSortedness


cdef class CoreAdvancedOrderbook:
    cdef:
        double _tick_size
        double _lot_size
        double _tick_size_recip  # Pre-computed 1.0 / tick_size for fast conversion
        double _lot_size_recip   # Pre-computed 1.0 / lot_size for fast conversion
        u64 _max_levels

        CyOrderbookSortedness _delta_sortedness
        CyOrderbookSortedness _snapshot_sortedness

        OrderbookLadder _bids
        OrderbookLadder _asks
        OrderbookLadderData* _bids_data
        OrderbookLadderData* _asks_data

    # Internal helpers
    cdef void _ensure_not_empty(self)
    cdef bint _check_if_empty(self)
    cdef void _normalize_incoming_levels(self, OrderbookLevels asks, OrderbookLevels bids, bint is_snapshot)
    cdef void _process_matching_ask_ticks(self, OrderbookLevel* ask)
    cdef void _process_matching_bid_ticks(self, OrderbookLevel* bid)
    cdef void _process_lower_ask_ticks(self, OrderbookLevel* ask)
    cdef void _process_higher_bid_ticks(self, OrderbookLevel* bid)
    cdef void _process_middle_ask_ticks(self, OrderbookLevel* ask)
    cdef void _process_middle_bid_ticks(self, OrderbookLevel* bid)
    cdef void _assign_bbo_level(self, OrderbookLevel* target, OrderbookLevel* source, u64 ticks, u64 lots) noexcept nogil

    cdef void clear(self) 
    cdef void consume_snapshot(self, OrderbookLevels new_asks, OrderbookLevels new_bids)
    cdef void consume_deltas(self, OrderbookLevels asks, OrderbookLevels bids)
    cdef void consume_bbo(self, OrderbookLevel ask, OrderbookLevel bid)

    cdef double get_mid_price(self)
    cdef double get_bbo_spread(self)
    cdef double get_wmid_price(self)
    cdef double get_volume_weighted_mid_price(self, double size, bint is_base_currency)
    cdef double get_price_impact(self, double size, bint is_buy, bint is_base_currency)
    cdef double get_size_for_price_impact_bps(self, double impact_bps, bint is_buy, bint is_base_currency)
    cdef bint is_bbo_crossed(self, double other_bid_price, double other_ask_price)
    cdef bint does_bbo_price_change(self, double bid_price, double ask_price)
    cdef OrderbookLadderData* get_bids_data(self) noexcept
    cdef OrderbookLadderData* get_asks_data(self) noexcept 



"""Tests for logarithmic weight calculations."""

import numpy as np
import pytest

from mm_toolbox.weights import logarithmic_weights


class TestLogarithmicWeightsBasic:
    """Test basic logarithmic weights functionality."""

    def test_function_return_type(self):
        """Test that logarithmic_weights returns proper numpy array type."""
        result = logarithmic_weights(5)
        assert isinstance(result, np.ndarray)
        assert result.dtype == np.float64

    def test_default_calculation(self):
        """Test logarithmic weights with default parameters."""
        result = logarithmic_weights(5)
        expected = np.array([0.0, 0.14478295, 0.22947555, 0.2895659, 0.3361756])
        np.testing.assert_allclose(result, expected, rtol=1e-6)

    def test_weights_normalization(self):
        """Test that weights sum to 1.0."""
        result = logarithmic_weights(5)
        assert pytest.approx(result.sum(), abs=1e-12) == 1.0

        result_large = logarithmic_weights(50)
        assert pytest.approx(result_large.sum(), abs=1e-12) == 1.0

    def test_weights_length(self):
        """Test that weights array has correct length."""
        for num in [3, 5, 10, 20]:
            result = logarithmic_weights(num)
            assert len(result) == num

    def test_weights_ordering(self):
        """Test that weights are in ascending order."""
        result = logarithmic_weights(5)
        # Logarithmic weights should be in ascending order (more recent = higher weight)
        for i in range(len(result) - 1):
            assert result[i] <= result[i + 1]

    def test_first_weight_is_zero(self):
        """Test that first weight is always zero."""
        for num in [3, 5, 10, 20]:
            result = logarithmic_weights(num)
            assert result[0] == pytest.approx(0.0, abs=1e-12)


class TestLogarithmicWeightsScaling:
    """Test logarithmic weights with different sizes."""

    def test_larger_num_values(self):
        """Test with larger number of weights."""
        result = logarithmic_weights(100)
        assert len(result) == 100
        assert pytest.approx(result.sum(), abs=1e-10) == 1.0

        # Should be monotonically increasing
        for i in range(len(result) - 1):
            assert result[i] <= result[i + 1]

        # First weight should be zero
        assert result[0] == pytest.approx(0.0, abs=1e-12)

    def test_different_num_sizes(self):
        """Test logarithmic weights with various sizes."""
        sizes = [3, 5, 10, 25, 50]

        for num in sizes:
            result = logarithmic_weights(num)

            # Basic properties
            assert len(result) == num
            assert pytest.approx(result.sum(), abs=1e-10) == 1.0
            assert result[0] == pytest.approx(0.0, abs=1e-12)

            # Monotonically increasing
            for i in range(len(result) - 1):
                assert result[i] <= result[i + 1]

    def test_scaling_behavior(self):
        """Test how weights scale with size."""
        result_5 = logarithmic_weights(5)
        result_10 = logarithmic_weights(10)
        result_20 = logarithmic_weights(20)

        # Larger arrays should have smaller individual increments
        # (since they need to fit more weights in the same log curve)
        max_weight_5 = result_5[-1]
        max_weight_10 = result_10[-1]
        max_weight_20 = result_20[-1]

        # All should sum to 1, but distribution should be different
        # Test that max weights are meaningfully different (not just floating
        # point precision)
        assert abs(max_weight_5 - max_weight_10) > 0.05
        assert abs(max_weight_10 - max_weight_20) > 0.02


class TestLogarithmicWeightsEdgeCases:
    """Test edge cases and error handling."""

    def test_invalid_num_values(self):
        """Test validation of num parameter."""
        # num <= 1 should raise ValueError
        with pytest.raises(ValueError):
            logarithmic_weights(1)

        with pytest.raises(ValueError):
            logarithmic_weights(0)

        with pytest.raises(ValueError):
            logarithmic_weights(-1)

    def test_minimum_valid_num(self):
        """Test minimum valid num size."""
        result = logarithmic_weights(2)
        assert len(result) == 2
        assert pytest.approx(result.sum(), abs=1e-12) == 1.0
        assert result[0] == pytest.approx(0.0, abs=1e-12)
        assert result[1] == pytest.approx(1.0, abs=1e-12)

    def test_small_num_values(self):
        """Test with small but valid num values."""
        # Test num = 3
        result_3 = logarithmic_weights(3)
        assert len(result_3) == 3
        assert pytest.approx(result_3.sum(), abs=1e-12) == 1.0
        assert result_3[0] == pytest.approx(0.0, abs=1e-12)
        assert result_3[1] <= result_3[2]

        # Test num = 4
        result_4 = logarithmic_weights(4)
        assert len(result_4) == 4
        assert pytest.approx(result_4.sum(), abs=1e-12) == 1.0
        assert result_4[0] == pytest.approx(0.0, abs=1e-12)


class TestLogarithmicWeightsNumerical:
    """Test numerical properties and stability."""

    def test_numerical_precision(self):
        """Test numerical precision with different sizes."""
        for num in [5, 10, 50, 100]:
            result = logarithmic_weights(num)
            # Sum should be very close to 1.0
            assert pytest.approx(result.sum(), abs=1e-10) == 1.0
            # All weights should be non-negative
            assert np.all(result >= 0)
            # First weight should be exactly zero
            assert result[0] == pytest.approx(0.0, abs=1e-15)

    def test_consistency_across_calls(self):
        """Test that repeated calls give consistent results."""
        num = 10
        result1 = logarithmic_weights(num)
        result2 = logarithmic_weights(num)
        result3 = logarithmic_weights(num)

        np.testing.assert_array_equal(result1, result2)
        np.testing.assert_array_equal(result2, result3)

    def test_mathematical_properties(self):
        """Test mathematical properties of logarithmic weights."""
        num = 8
        result = logarithmic_weights(num)

        # First weight should be exactly zero
        assert result[0] == 0.0

        # All other weights should be positive
        assert np.all(result[1:] > 0)

        # Should be monotonically increasing
        for i in range(len(result) - 1):
            assert result[i] <= result[i + 1]

    def test_logarithmic_curve_properties(self):
        """Test that weights follow logarithmic curve properties."""
        num = 10
        result = logarithmic_weights(num)

        # Calculate differences between consecutive weights
        diffs = np.diff(result)

        # Differences should generally decrease (logarithmic curve flattens)
        # Allow some tolerance for numerical precision
        decreasing_count = 0
        for i in range(len(diffs) - 1):
            if diffs[i] >= diffs[i + 1]:
                decreasing_count += 1

        # Most differences should be decreasing (characteristic of log curve)
        assert decreasing_count >= len(diffs) // 2

    def test_weight_distribution_characteristics(self):
        """Test characteristics of weight distribution."""
        result = logarithmic_weights(10)

        # Most weight should be concentrated in later positions
        first_half_sum = result[:5].sum()
        second_half_sum = result[5:].sum()
        assert second_half_sum > first_half_sum

        # Last weight should be the largest
        assert result[-1] == np.max(result)

        # First weight should be the smallest (zero)
        assert result[0] == np.min(result)


class TestLogarithmicWeightsComparison:
    """Test comparisons between different logarithmic weight configurations."""

    def test_relative_weight_distribution(self):
        """Test relative distribution of weights."""
        small = logarithmic_weights(5)
        large = logarithmic_weights(10)

        # Both should sum to 1
        assert pytest.approx(small.sum(), abs=1e-12) == 1.0
        assert pytest.approx(large.sum(), abs=1e-12) == 1.0

        # Both should start with zero
        assert small[0] == pytest.approx(0.0, abs=1e-12)
        assert large[0] == pytest.approx(0.0, abs=1e-12)

        # Both should be monotonically increasing
        assert np.all(np.diff(small) >= 0)
        assert np.all(np.diff(large) >= 0)

    def test_convergence_properties(self):
        """Test convergence properties with increasing size."""
        sizes = [5, 10, 20, 50]
        max_weights = []

        for size in sizes:
            result = logarithmic_weights(size)
            max_weights.append(result[-1])

        # Maximum weights should generally decrease as size increases
        # (more weights to distribute among)
        for i in range(len(max_weights) - 1):
            assert max_weights[i] >= max_weights[i + 1] * 0.8  # Allow some tolerance



"""
Orderbook sortedness enums for level array ordering.

Defines PyOrderbookSortedness (Python IntEnum) and CyOrderbookSortedness (Cython
enum) with conversion utilities. Used to specify expected sort order of incoming
snapshot/delta levels to optimize or skip sorting operations.
"""
from __future__ import annotations

from enum import IntEnum


class PyOrderbookSortedness(IntEnum):
    UNKNOWN = 0
    ASCENDING = 1
    DESCENDING = 2
    BIDS_ASCENDING_ASKS_DESCENDING = 3
    BIDS_DESCENDING_ASKS_ASCENDING = 4

cpdef CyOrderbookSortedness py_to_cy_orderbook_sortedness(object sortedness):
    """Convert a Python-facing sortedness enum to a Cython-facing sortedness enum."""
    if sortedness == PyOrderbookSortedness.UNKNOWN:
        return CyOrderbookSortedness.UNKNOWN
    elif sortedness == PyOrderbookSortedness.ASCENDING:
        return CyOrderbookSortedness.ASCENDING
    elif sortedness == PyOrderbookSortedness.DESCENDING:
        return CyOrderbookSortedness.DESCENDING
    elif sortedness == PyOrderbookSortedness.BIDS_ASCENDING_ASKS_DESCENDING:
        return CyOrderbookSortedness.BIDS_ASCENDING_ASKS_DESCENDING
    elif sortedness == PyOrderbookSortedness.BIDS_DESCENDING_ASKS_ASCENDING:
        return CyOrderbookSortedness.BIDS_DESCENDING_ASKS_ASCENDING
    else:
        return CyOrderbookSortedness.UNKNOWN



"""Type stubs for config module."""

from mm_toolbox.logging.advanced.pylog import PyLogLevel

class LoggerConfig:
    """Configuration class for the logging system."""

    base_level: int
    do_stdout: bool
    str_format: str
    path: str
    flush_interval_s: float
    emit_internal: bool
    ipc_linger_ms: int

    def __init__(
        self,
        base_level: PyLogLevel | None = None,
        do_stdout: bool = False,
        str_format: str = "%(asctime)s [%(levelname)s] %(name)s - %(message)s",
        path: str = "ipc:///tmp/hft_logger",
        flush_interval_s: float = 1.0,
        emit_internal: bool = False,
        ipc_linger_ms: int = 1000,
    ) -> None:
        """Initialize the LoggerConfig with transport, path, and format settings.

        Args:
            base_level: The minimum log level to process. Defaults to PyLogLevel.INFO.
            do_stdout: Whether to print to stdout. Defaults to False.
            str_format: The format string for log messages, similar to Python's
                default logger. Supports standard format placeholders like
                %(asctime)s, %(name)s, %(levelname)s, and %(message)s.
                Defaults to '%(asctime)s [%(levelname)s] %(name)s - %(message)s'.
                Must contain at least %(message)s to be valid.
            path: The connection path for the transport protocol. Must follow the format
                required by the selected transport protocol. Defaults to 'ipc:///tmp/hft_logger'.
            flush_interval_s: Timeout in seconds for log messages. Defaults to 1.0.
            emit_internal: If True, emit internal startup/shutdown logs. Defaults to False.
            ipc_linger_ms: ZMQ linger in milliseconds for IPC sockets. Defaults to 1000.
        """
        ...



import os
import platform
import sys

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext as _build_ext

"""
# Usage instructions
# 
# To build
#   'python setup.py build_ext --inplace'
#   'python setup.py build_ext --inplace --parallel 4'
# 
# To install the package
#   'pip install .'
"""

COMPILER_DIRECTIVES = {
    "language_level": 3,
    "boundscheck": False,
    "wraparound": False,
    "embedsignature": True,
    "embedsignature.format": "python",
    "cdivision": True,
    "cpow": True,
}
EXTRA_COMPILE_ARGS = [
    "-O3",
    "-Wno-sign-compare",
    "-Wno-unused-function",
    "-Wno-unreachable-code",
]

EXTRA_LINK_ARGS = ["-Wl,-w"] if sys.platform == "darwin" else []


def get_extension(name, sources, include_dirs=None):
    """Helper to create extension with consistent settings."""
    return Extension(
        name=name,
        sources=sources,
        include_dirs=["src"] if not include_dirs else include_dirs,
        extra_compile_args=EXTRA_COMPILE_ARGS,
        extra_link_args=EXTRA_LINK_ARGS,
    )


def get_build_dir():
    # Example: build/cython.linux-x86_64-cpython-313
    plat = platform.system().lower()
    machine = platform.machine().lower()
    impl = platform.python_implementation().lower()
    # e.g. cpython-313
    py_tag = f"{impl}-{sys.version_info.major}{sys.version_info.minor}{sys.version_info.micro}"
    return f"build/cython.{plat}-{machine}-{py_tag}"


#   * Hierarchy of build dependencies *
#
#   - Rounding (no dependencies)
#   - Time (no dependencies)
#   - Ringbuffer (no dependencies)
#   - Orderbook (no dependencies)
#
#   - Moving Average (depends on Ringbuffer & Time)
#   - Candles (depends on Ringbuffer & Time)
#   - Logging (depends on Ringbuffer & Time)
#   - Websocket (depends on Ringbuffer, Moving Average, & Time)
#   - Misc.Filter (no dependencies) / Misc.Limiter (depends on Time)


def get_rounding_extensions():
    """Get list of rounding extensions for compilation."""
    return [
        get_extension(
            name="mm_toolbox.rounding.rounder",
            sources=["src/mm_toolbox/rounding/rounder.pyx"],
            include_dirs=[np.get_include(), "src"],
        ),
    ]


def get_time_extensions():
    """Get list of time extensions for compilation."""
    return [
        get_extension(
            name="mm_toolbox.time.time",
            sources=[
                "src/mm_toolbox/time/ctime_impl.c",
                "src/mm_toolbox/time/time.pyx",
            ],
            include_dirs=["src/mm_toolbox/time", "src"],
        ),
    ]


def get_ringbuffer_extensions():
    """Get list of ringbuffer extensions for compilation."""
    return [
        get_extension(
            name="mm_toolbox.ringbuffer.generic",
            sources=["src/mm_toolbox/ringbuffer/generic.pyx"],
            include_dirs=[np.get_include(), "src/mm_toolbox/ringbuffer", "src"],
        ),
        get_extension(
            name="mm_toolbox.ringbuffer.numeric",
            sources=["src/mm_toolbox/ringbuffer/numeric.pyx"],
            include_dirs=[np.get_include(), "src/mm_toolbox/ringbuffer", "src"],
        ),
        get_extension(
            name="mm_toolbox.ringbuffer.bytes",
            sources=["src/mm_toolbox/ringbuffer/bytes.pyx"],
            include_dirs=[np.get_include(), "src/mm_toolbox/ringbuffer", "src"],
        ),
        # SHM ringbuffer (shared memory SPSC queue)
        get_extension(
            name="mm_toolbox.ringbuffer.shm.core",
            sources=[
                "src/mm_toolbox/time/ctime_impl.c",
                "src/mm_toolbox/ringbuffer/shm/c/shm_helpers.c",
                "src/mm_toolbox/ringbuffer/shm/c/shm_core.c",
                "src/mm_toolbox/ringbuffer/shm/core.pyx",
            ],
            include_dirs=[
                np.get_include(),
                "src/mm_toolbox/ringbuffer/shm",
                "src/mm_toolbox/ringbuffer/shm/c",
                "src/mm_toolbox/ringbuffer",
                "src/mm_toolbox/time",
                "src",
            ],
        ),
        get_extension(
            name="mm_toolbox.ringbuffer.shm.memory",
            sources=["src/mm_toolbox/ringbuffer/shm/memory.pyx"],
            include_dirs=[
                np.get_include(),
                "src/mm_toolbox/ringbuffer/shm",
                "src",
            ],
        ),
        get_extension(
            name="mm_toolbox.ringbuffer.shm.atomics",
            sources=["src/mm_toolbox/ringbuffer/shm/atomics.pyx"],
            include_dirs=[
                np.get_include(),
                "src/mm_toolbox/ringbuffer/shm",
                "src",
            ],
        ),
    ]


def get_orderbook_extensions():
    """Get list of orderbook extensions for compilation."""
    base_include_dirs = [
        np.get_include(),
        "src/mm_toolbox/orderbook/advanced",
        "src/mm_toolbox/orderbook/advanced/c",
        "src/mm_toolbox/orderbook/advanced/enum",
        "src/mm_toolbox/orderbook/advanced/level",
        "src/mm_toolbox/orderbook/advanced/ladder",
        "src",
    ]
    return [
        get_extension(
            name="mm_toolbox.orderbook.advanced.enum.enums",
            sources=["src/mm_toolbox/orderbook/advanced/enum/enums.pyx"],
            include_dirs=base_include_dirs,
        ),
        get_extension(
            name="mm_toolbox.orderbook.advanced.level.level",
            sources=["src/mm_toolbox/orderbook/advanced/level/level.pyx"],
            include_dirs=base_include_dirs,
        ),
        get_extension(
            name="mm_toolbox.orderbook.advanced.level.helpers",
            sources=[
                "src/mm_toolbox/orderbook/advanced/c/orderbook_helpers.c",
                "src/mm_toolbox/orderbook/advanced/level/helpers.pyx",
            ],
            include_dirs=base_include_dirs,
        ),
        get_extension(
            name="mm_toolbox.orderbook.advanced.ladder.ladder",
            sources=[
                "src/mm_toolbox/orderbook/advanced/c/orderbook_ladder.c",
                "src/mm_toolbox/orderbook/advanced/ladder/ladder.pyx",
            ],
            include_dirs=base_include_dirs,
        ),
        get_extension(
            name="mm_toolbox.orderbook.advanced.core",
            sources=["src/mm_toolbox/orderbook/advanced/core.pyx"],
            include_dirs=base_include_dirs,
        ),
        get_extension(
            name="mm_toolbox.orderbook.advanced.cython",
            sources=["src/mm_toolbox/orderbook/advanced/cython.pyx"],
            include_dirs=base_include_dirs,
        ),
        get_extension(
            name="mm_toolbox.orderbook.advanced.python",
            sources=["src/mm_toolbox/orderbook/advanced/python.pyx"],
            include_dirs=base_include_dirs,
        ),
    ]


def get_logging_extensions():
    """Get list of logging extensions for compilation."""
    return [
        get_extension(
            name="mm_toolbox.logging.advanced.protocol",
            sources=["src/mm_toolbox/logging/advanced/protocol.pyx"],
            include_dirs=["src/mm_toolbox/logging/advanced", "src"],
        ),
        get_extension(
            name="mm_toolbox.logging.advanced.config",
            sources=["src/mm_toolbox/logging/advanced/config.pyx"],
            include_dirs=["src/mm_toolbox/logging/advanced", "src"],
        ),
        get_extension(
            name="mm_toolbox.logging.advanced.master",
            sources=["src/mm_toolbox/logging/advanced/master.pyx"],
            include_dirs=["src/mm_toolbox/logging/advanced", "src"],
        ),
        get_extension(
            name="mm_toolbox.logging.advanced.worker",
            sources=["src/mm_toolbox/logging/advanced/worker.pyx"],
            include_dirs=["src/mm_toolbox/logging/advanced", "src"],
        ),
    ]


def get_moving_average_extensions():
    """Get list of moving average extensions for compilation."""
    return [
        get_extension(
            name="mm_toolbox.moving_average.base",
            sources=["src/mm_toolbox/moving_average/base.pyx"],
            include_dirs=[np.get_include(), "src/mm_toolbox/moving_average", "src"],
        ),
        get_extension(
            name="mm_toolbox.moving_average.ema",
            sources=["src/mm_toolbox/moving_average/ema.pyx"],
            include_dirs=[np.get_include(), "src/mm_toolbox/moving_average", "src"],
        ),
        get_extension(
            name="mm_toolbox.moving_average.sma",
            sources=["src/mm_toolbox/moving_average/sma.pyx"],
            include_dirs=[np.get_include(), "src/mm_toolbox/moving_average", "src"],
        ),
        get_extension(
            name="mm_toolbox.moving_average.tema",
            sources=["src/mm_toolbox/moving_average/tema.pyx"],
            include_dirs=[np.get_include(), "src/mm_toolbox/moving_average", "src"],
        ),
        get_extension(
            name="mm_toolbox.moving_average.wma",
            sources=["src/mm_toolbox/moving_average/wma.pyx"],
            include_dirs=[np.get_include(), "src/mm_toolbox/moving_average", "src"],
        ),
    ]


def get_candles_extensions():
    """Get list of candles extensions for compilation."""
    return [
        get_extension(
            name="mm_toolbox.candles.base",
            sources=["src/mm_toolbox/candles/base.pyx"],
            include_dirs=[np.get_include(), "src"],
        ),
        get_extension(
            name="mm_toolbox.candles.multi",
            sources=["src/mm_toolbox/candles/multi.pyx"],
            include_dirs=[np.get_include(), "src"],
        ),
        get_extension(
            name="mm_toolbox.candles.time",
            sources=["src/mm_toolbox/candles/time.pyx"],
            include_dirs=[np.get_include(), "src"],
        ),
        get_extension(
            name="mm_toolbox.candles.tick",
            sources=["src/mm_toolbox/candles/tick.pyx"],
            include_dirs=[np.get_include(), "src"],
        ),
        get_extension(
            name="mm_toolbox.candles.volume",
            sources=["src/mm_toolbox/candles/volume.pyx"],
            include_dirs=[np.get_include(), "src"],
        ),
        get_extension(
            name="mm_toolbox.candles.price",
            sources=["src/mm_toolbox/candles/price.pyx"],
            include_dirs=[np.get_include(), "src"],
        ),
    ]


def get_websocket_extensions():
    """Get list of websocket extensions for compilation."""
    return [
        get_extension(
            name="mm_toolbox.websocket.connection",
            sources=["src/mm_toolbox/websocket/connection.pyx"],
            include_dirs=[np.get_include(), "src"],
        ),
    ]


def get_misc_extensions():
    """Get list of misc extensions for compilation."""
    return [
        get_extension(
            name="mm_toolbox.misc.filter.core",
            sources=["src/mm_toolbox/misc/filter/core.pyx"],
            include_dirs=["src", "src/mm_toolbox/misc/filter"],
        ),
    ]


def get_rate_limiter_extensions():
    """Get list of rate limiter extensions for compilation."""
    base_include_dirs = ["src", "src/mm_toolbox/rate_limiter"]
    return [
        get_extension(
            name="mm_toolbox.rate_limiter.result",
            sources=["src/mm_toolbox/rate_limiter/result.pyx"],
            include_dirs=base_include_dirs,
        ),
        get_extension(
            name="mm_toolbox.rate_limiter.state",
            sources=["src/mm_toolbox/rate_limiter/state.pyx"],
            include_dirs=base_include_dirs,
        ),
        get_extension(
            name="mm_toolbox.rate_limiter.bucket",
            sources=["src/mm_toolbox/rate_limiter/bucket.pyx"],
            include_dirs=base_include_dirs,
        ),
        get_extension(
            name="mm_toolbox.rate_limiter.limiter",
            sources=["src/mm_toolbox/rate_limiter/limiter.pyx"],
            include_dirs=base_include_dirs,
        ),
    ]


module_list: list[Extension] = []
module_list.extend(get_rounding_extensions())
module_list.extend(get_time_extensions())
module_list.extend(get_ringbuffer_extensions())
module_list.extend(get_orderbook_extensions())
module_list.extend(get_moving_average_extensions())
module_list.extend(get_candles_extensions())
module_list.extend(get_logging_extensions())
module_list.extend(get_websocket_extensions())
module_list.extend(get_misc_extensions())
module_list.extend(get_rate_limiter_extensions())


class build_ext(_build_ext):
    """Custom build_ext to remove generated .c files after build."""

    def run(self):
        """Run the build_ext command and clean up generated C files."""
        super().run()
        # Remove all generated .c files from cythonized .pyx sources
        for ext in self.extensions:
            for src in ext.sources:
                if src.endswith(".pyx"):
                    c_file = src.replace(".pyx", ".c")
                    if os.path.exists(c_file):
                        try:
                            os.remove(c_file)
                        except Exception:
                            pass


setup(
    cmdclass={"build_ext": build_ext},
    name="mm_toolbox",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={
        "mm_toolbox": [
            "**/*.pxd",
            "**/*.pyi",
        ]
    },
    ext_modules=cythonize(
        module_list=module_list,
        compiler_directives=COMPILER_DIRECTIVES,
        build_dir=get_build_dir(),
        include_path=[
            "src",
            "src/mm_toolbox/misc/filter",
            "src/mm_toolbox/rate_limiter",
        ],
    ),
)



"""Tests for Rounder implementation."""

import numpy as np
import pytest

from mm_toolbox.rounding import Rounder
from mm_toolbox.rounding.rounder import RounderConfig


class TestRounderConfig:
    """Test RounderConfig validation comprehensively."""

    def test_valid_config_creation(self):
        """Test creating valid configurations."""
        cfg = RounderConfig(
            tick_size=0.01,
            lot_size=0.001,
            round_bids_down=True,
            round_asks_up=True,
            round_size_up=True,
        )
        assert cfg.tick_size == 0.01
        assert cfg.lot_size == 0.001
        assert cfg.round_bids_down is True
        assert cfg.round_asks_up is True
        assert cfg.round_size_up is True

    def test_tick_size_validation(self):
        """Test tick_size validation."""
        # Valid tick sizes
        valid_tick_sizes = [0.01, 0.001, 0.0001, 1.0, 0.5]
        for tick_size in valid_tick_sizes:
            cfg = RounderConfig(
                tick_size=tick_size,
                lot_size=0.001,
                round_bids_down=True,
                round_asks_up=True,
                round_size_up=True,
            )
            assert cfg.tick_size == tick_size

        # Invalid tick sizes
        invalid_tick_sizes = [0.0, -0.01, -1.0]
        for tick_size in invalid_tick_sizes:
            with pytest.raises(
                ValueError, match="Invalid tick_size; must be greater than 0"
            ):
                RounderConfig(
                    tick_size=tick_size,
                    lot_size=0.001,
                    round_bids_down=True,
                    round_asks_up=True,
                    round_size_up=True,
                )

    def test_lot_size_validation(self):
        """Test lot_size validation."""
        # Valid lot sizes
        valid_lot_sizes = [0.01, 0.001, 0.0001, 1.0, 0.5]
        for lot_size in valid_lot_sizes:
            cfg = RounderConfig(
                tick_size=0.01,
                lot_size=lot_size,
                round_bids_down=True,
                round_asks_up=True,
                round_size_up=True,
            )
            assert cfg.lot_size == lot_size

        # Invalid lot sizes
        invalid_lot_sizes = [0.0, -0.001, -1.0]
        for lot_size in invalid_lot_sizes:
            with pytest.raises(
                ValueError, match="Invalid lot_size; must be greater than 0"
            ):
                RounderConfig(
                    tick_size=0.01,
                    lot_size=lot_size,
                    round_bids_down=True,
                    round_asks_up=True,
                    round_size_up=True,
                )

    def test_default_config_creation(self):
        """Test default configuration creation."""
        cfg = RounderConfig.default(tick_size=0.01, lot_size=0.001)
        assert cfg.tick_size == 0.01
        assert cfg.lot_size == 0.001
        assert cfg.round_bids_down is True
        assert cfg.round_asks_up is True
        assert cfg.round_size_up is True

    def test_custom_rounding_configurations(self):
        """Test different rounding direction configurations."""
        # Custom configuration: round bids up, asks down, sizes down
        cfg = RounderConfig(
            tick_size=0.01,
            lot_size=0.001,
            round_bids_down=False,
            round_asks_up=False,
            round_size_up=False,
        )
        assert cfg.round_bids_down is False
        assert cfg.round_asks_up is False
        assert cfg.round_size_up is False


class TestRounderBasicOperations:
    """Test basic Rounder operations."""

    def test_rounder_creation(self):
        """Test creating Rounder instances."""
        cfg = RounderConfig.default(0.01, 0.001)
        rounder = Rounder(cfg)

        assert rounder is not None
        assert isinstance(rounder.bid(1.234), float)
        assert isinstance(rounder.ask(1.234), float)
        assert isinstance(rounder.size(1.234), float)

    def test_single_bid_rounding_default(self):
        """Test single bid rounding with default configuration (round down)."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))

        # Test exact tick multiples
        assert rounder.bid(1.00) == 1.00
        assert rounder.bid(1.01) == 1.01
        assert rounder.bid(1.02) == 1.02

        # Test values between ticks (should round down)
        assert rounder.bid(1.005) == 1.00
        assert rounder.bid(1.014) == 1.01
        assert rounder.bid(1.025) == 1.02

        # Test edge cases
        assert rounder.bid(1.009) == 1.00
        assert rounder.bid(1.019) == 1.01

    def test_single_ask_rounding_default(self):
        """Test single ask rounding with default configuration (round up)."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))

        # Test exact tick multiples
        assert rounder.ask(1.00) == 1.00
        assert rounder.ask(1.01) == 1.01
        assert rounder.ask(1.02) == 1.02

        # Test values between ticks (should round up)
        assert rounder.ask(1.005) == 1.01
        assert rounder.ask(1.014) == 1.02
        assert rounder.ask(1.025) == 1.03

    def test_single_size_rounding_default(self):
        """Test single size rounding with default configuration (round up)."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))

        # Test exact lot multiples
        assert rounder.size(1.000) == 1.000
        assert rounder.size(1.001) == 1.001
        assert rounder.size(1.002) == 1.002

        # Test values between lots (should round up)
        assert rounder.size(1.0005) == 1.001
        assert rounder.size(1.0014) == 1.002
        assert rounder.size(1.0025) == 1.003

    def test_array_bids_basic(self):
        """Test basic array bid rounding."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))
        prices = np.array([1.234, 1.231, 1.237, 1.240, 1.245])
        expected = np.array([1.23, 1.23, 1.23, 1.24, 1.24])

        result = rounder.bids(prices)
        np.testing.assert_allclose(result, expected, rtol=0, atol=1e-12)
        assert result.dtype == np.float64

    def test_array_asks_basic(self):
        """Test basic array ask rounding."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))
        prices = np.array([1.234, 1.231, 1.237, 1.240, 1.245])
        expected = np.array([1.24, 1.24, 1.24, 1.24, 1.25])

        result = rounder.asks(prices)
        np.testing.assert_allclose(result, expected, rtol=0, atol=1e-12)
        assert result.dtype == np.float64

    def test_array_sizes_basic(self):
        """Test basic array size rounding."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))
        sizes = np.array([1.234, 1.231, 1.237, 1.240, 1.245])
        expected = np.array([1.234, 1.231, 1.237, 1.240, 1.245])

        result = rounder.sizes(sizes)
        np.testing.assert_allclose(result, expected, rtol=0, atol=1e-12)
        assert result.dtype == np.float64


class TestRounderCustomConfigurations:
    """Test Rounder with custom configurations."""

    def test_custom_bid_rounding(self):
        """Test bid rounding with custom configuration (round up)."""
        cfg = RounderConfig(
            tick_size=0.01,
            lot_size=0.001,
            round_bids_down=False,
            round_asks_up=True,
            round_size_up=True,
        )
        rounder = Rounder(cfg)

        # Test values between ticks (should round up)
        assert rounder.bid(1.005) == 1.01
        assert rounder.bid(1.014) == 1.02
        assert rounder.bid(1.025) == 1.03

    def test_custom_ask_rounding(self):
        """Test ask rounding with custom configuration (round down)."""
        cfg = RounderConfig(
            tick_size=0.01,
            lot_size=0.001,
            round_bids_down=True,
            round_asks_up=False,
            round_size_up=True,
        )
        rounder = Rounder(cfg)

        # Test values between ticks (should round down)
        assert rounder.ask(1.005) == 1.00
        assert rounder.ask(1.014) == 1.01
        assert rounder.ask(1.025) == 1.02

    def test_custom_size_rounding(self):
        """Test size rounding with custom configuration (round down)."""
        cfg = RounderConfig(
            tick_size=0.01,
            lot_size=0.001,
            round_bids_down=True,
            round_asks_up=True,
            round_size_up=False,
        )
        rounder = Rounder(cfg)

        # Test values between lots (should round down)
        assert rounder.size(1.0005) == 1.000
        assert rounder.size(1.0014) == 1.001
        assert rounder.size(1.0025) == 1.002

    def test_array_operations_custom_config(self):
        """Test array operations with custom configurations."""
        cfg = RounderConfig(
            tick_size=0.01,
            lot_size=0.001,
            round_bids_down=False,
            round_asks_up=False,
            round_size_up=False,
        )
        rounder = Rounder(cfg)

        prices = np.array([1.234, 1.231, 1.237, 1.240, 1.245])

        # Bids should round up
        expected_bids = np.array([1.24, 1.24, 1.24, 1.24, 1.25])
        result_bids = rounder.bids(prices)
        np.testing.assert_allclose(result_bids, expected_bids, rtol=0, atol=1e-12)

        # Asks should round down
        expected_asks = np.array([1.23, 1.23, 1.23, 1.24, 1.24])
        result_asks = rounder.asks(prices)
        np.testing.assert_allclose(result_asks, expected_asks, rtol=0, atol=1e-12)

    def test_rounding_direction_consistency(self):
        """Test that rounding direction is consistent with configuration."""
        default_rounder = Rounder(RounderConfig.default(0.01, 0.001))
        custom_cfg = RounderConfig(
            tick_size=0.01,
            lot_size=0.001,
            round_bids_down=False,
            round_asks_up=False,
            round_size_up=False,
        )
        custom_rounder = Rounder(custom_cfg)

        test_price = 1.005

        # Default: bids down, asks up
        assert default_rounder.bid(test_price) < test_price
        assert default_rounder.ask(test_price) > test_price

        # Custom: bids up, asks down
        assert custom_rounder.bid(test_price) > test_price
        assert custom_rounder.ask(test_price) < test_price


class TestRounderDifferentTickSizes:
    """Test Rounder with different tick and lot sizes."""

    def test_small_tick_precision(self):
        """Test precision with very small tick sizes."""
        rounder = Rounder(RounderConfig.default(0.0001, 0.00001))

        # Test 0.0001 tick size
        assert rounder.bid(1.12345) == 1.1234
        assert rounder.ask(1.12345) == 1.1235
        assert rounder.bid(1.12346) == 1.1234
        assert rounder.ask(1.12346) == 1.1235

    def test_large_tick_rounding(self):
        """Test rounding with large tick sizes."""
        rounder = Rounder(RounderConfig.default(1.0, 0.1))

        # Test 1.0 tick size
        assert rounder.bid(1.4) == 1.0
        assert rounder.ask(1.4) == 2.0
        assert rounder.bid(1.6) == 1.0
        assert rounder.ask(1.6) == 2.0

    def test_fractional_tick_sizes(self):
        """Test with fractional tick sizes."""
        rounder = Rounder(RounderConfig.default(0.5, 0.1))

        # Test 0.5 tick size (simpler fractions to avoid precision issues)
        # Valid multiples of 0.5: 0.0, 0.5, 1.0, 1.5, 2.0, etc.

        # Test basic rounding behavior
        assert rounder.bid(1.2) == 1.0  # 1.2 rounds down to 1.0
        assert rounder.ask(1.2) == 1.5  # 1.2 rounds up to 1.5

        assert rounder.bid(1.7) == 1.5  # 1.7 rounds down to 1.5
        assert rounder.ask(1.7) == 2.0  # 1.7 rounds up to 2.0

        # Test exact multiples remain unchanged
        assert rounder.bid(1.0) == 1.0
        assert rounder.ask(1.0) == 1.0
        assert rounder.bid(1.5) == 1.5
        assert rounder.ask(1.5) == 1.5

    def test_lot_size_variations(self):
        """Test with different lot sizes."""
        rounder = Rounder(RounderConfig.default(0.01, 0.01))

        # Test 0.01 lot size
        assert rounder.size(1.234) == 1.24
        assert rounder.size(1.235) == 1.24
        assert rounder.size(1.236) == 1.24


class TestRounderArrayOperations:
    """Test array operations comprehensively."""

    def test_empty_arrays(self):
        """Test handling of empty arrays."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))
        empty_prices = np.array([])
        empty_sizes = np.array([])

        result_bids = rounder.bids(empty_prices)
        result_asks = rounder.asks(empty_prices)
        result_sizes = rounder.sizes(empty_sizes)

        assert len(result_bids) == 0
        assert len(result_asks) == 0
        assert len(result_sizes) == 0
        assert result_bids.dtype == np.float64
        assert result_asks.dtype == np.float64
        assert result_sizes.dtype == np.float64

    def test_single_element_arrays(self):
        """Test arrays with single elements."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))
        single_price = np.array([1.234])
        single_size = np.array([1.234])

        result_bid = rounder.bids(single_price)
        result_ask = rounder.asks(single_price)
        result_size = rounder.sizes(single_size)

        assert len(result_bid) == 1
        assert len(result_ask) == 1
        assert len(result_size) == 1
        assert result_bid[0] == 1.23
        assert result_ask[0] == 1.24
        assert result_size[0] == 1.234

    def test_large_arrays(self):
        """Test with reasonably large arrays."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))
        large_array = np.random.uniform(1.0, 2.0, 1000)

        result_bids = rounder.bids(large_array)
        result_asks = rounder.asks(large_array)
        result_sizes = rounder.sizes(large_array)

        assert len(result_bids) == len(large_array)
        assert len(result_asks) == len(large_array)
        assert len(result_sizes) == len(large_array)
        assert result_bids.dtype == np.float64
        assert result_asks.dtype == np.float64
        assert result_sizes.dtype == np.float64

    def test_different_dtypes(self):
        """Test handling of different numpy dtypes."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))

        # The Cython implementation expects double precision, so we test conversion
        prices_float32 = np.array([1.234, 1.231], dtype=np.float32)
        prices_float64 = np.array([1.234, 1.231], dtype=np.float64)

        # Convert float32 to float64 before passing to Cython methods
        result_32 = rounder.bids(prices_float32.astype(np.float64))
        result_64 = rounder.bids(prices_float64)

        assert result_32.dtype == np.float64
        assert result_64.dtype == np.float64
        np.testing.assert_allclose(result_32, result_64, rtol=0, atol=1e-12)

    def test_scalar_array_consistency(self):
        """Test that scalar and array methods produce consistent results."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))
        prices = np.array([1.234, 1.231, 1.237])

        # Test bid consistency
        scalar_bids = [rounder.bid(p) for p in prices]
        array_bids = rounder.bids(prices)
        np.testing.assert_allclose(scalar_bids, array_bids, rtol=0, atol=1e-12)

        # Test ask consistency
        scalar_asks = [rounder.ask(p) for p in prices]
        array_asks = rounder.asks(prices)
        np.testing.assert_allclose(scalar_asks, array_asks, rtol=0, atol=1e-12)

        # Test size consistency
        sizes = np.array([1.234, 1.231, 1.237])
        scalar_sizes = [rounder.size(s) for s in sizes]
        array_sizes = rounder.sizes(sizes)
        np.testing.assert_allclose(scalar_sizes, array_sizes, rtol=0, atol=1e-12)


class TestRounderEdgeCases:
    """Test edge cases and error handling."""

    def test_zero_and_negative_values(self):
        """Test handling of zero and negative values."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))

        # Zero values
        assert rounder.bid(0.0) == 0.0
        assert rounder.ask(0.0) == 0.0
        assert rounder.size(0.0) == 0.0

        # Negative values
        assert rounder.bid(-1.234) == -1.24
        assert rounder.ask(-1.234) == -1.23
        assert rounder.size(-1.234) == -1.234

    def test_extreme_values(self):
        """Test handling of extreme values."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))

        # Very large values
        large_price = 1e15
        assert rounder.bid(large_price) == pytest.approx(large_price, rel=1e-10)
        assert rounder.ask(large_price) == pytest.approx(large_price, rel=1e-10)

        # Very small values
        small_price = 1e-15
        assert rounder.bid(small_price) == 0.0
        assert rounder.ask(small_price) == 0.01

    def test_boundary_values(self):
        """Test values exactly at tick boundaries."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))

        # Test values exactly at tick boundaries
        assert rounder.bid(1.00) == 1.00
        assert rounder.ask(1.00) == 1.00

        # Test values just below and above tick boundaries
        assert rounder.bid(0.999) == 0.99
        assert rounder.ask(0.999) == 1.00
        assert rounder.bid(1.001) == 1.00
        assert rounder.ask(1.001) == 1.01

    def test_lot_size_boundaries(self):
        """Test lot size rounding edge cases."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))

        # Test values exactly at lot boundaries
        assert rounder.size(1.000) == 1.000
        assert rounder.size(1.001) == 1.001

        # Test values just below and above lot boundaries
        assert rounder.size(0.999) == 0.999
        assert rounder.size(1.0005) == 1.001
        assert rounder.size(1.0015) == 1.002

    def test_numerical_stability(self):
        """Test numerical stability with floating point arithmetic."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))

        # Test that repeated operations don't accumulate errors
        price = 1.234
        result1 = rounder.bid(price)
        result2 = rounder.bid(result1)
        result3 = rounder.bid(result2)

        # Results should be stable after first rounding
        assert result1 == result2
        assert result2 == result3

        # Test with more reasonable tick sizes for numerical stability
        medium_rounder = Rounder(RounderConfig.default(1e-6, 1e-6))
        medium_price = 1.0 + 1e-5
        result = medium_rounder.bid(medium_price)
        # With 1e-6 tick size, 1.00001 should round down to 1.00001 (10 ticks
        # up from 1.0)
        assert result == pytest.approx(1.00001, abs=1e-6)

    def test_memory_efficiency(self):
        """Test that array operations don't create unnecessary copies."""
        rounder = Rounder(RounderConfig.default(0.01, 0.001))
        prices = np.array([1.234, 1.231, 1.237])

        # Test that input array is not modified
        prices_original = prices.copy()
        result = rounder.bids(prices)
        np.testing.assert_array_equal(prices, prices_original)

        # Test that result is a new array
        assert result is not prices
        assert result.base is not prices.base if prices.base is not None else True

    def test_precision_edge_cases(self):
        """Test precision edge cases with different tick sizes."""
        # Test with simpler tick size to avoid floating point precision issues
        rounder = Rounder(RounderConfig.default(0.01, 0.001))

        # Test values that should round cleanly
        test_values = [1.005, 1.015, 1.025, 1.035]
        for value in test_values:
            bid_result = rounder.bid(value)
            ask_result = rounder.ask(value)

            # Results should be valid multiples of tick size (with some
            # tolerance for floating point)
            tick_multiple_bid = round(bid_result / 0.01)
            tick_multiple_ask = round(ask_result / 0.01)

            assert abs(bid_result - tick_multiple_bid * 0.01) < 1e-12
            assert abs(ask_result - tick_multiple_ask * 0.01) < 1e-12



"""Tests for standard logger configuration."""

import pytest

from mm_toolbox.logging.standard import LoggerConfig


class TestLoggerConfig:
    """Validate LoggerConfig inputs."""

    def test_default_values(self) -> None:
        cfg = LoggerConfig()
        assert cfg.base_level.name == "INFO"
        assert cfg.do_stdout is True
        assert cfg.flush_interval_s == 1.0
        assert cfg.buffer_size == 10000

    def test_invalid_flush_interval(self) -> None:
        with pytest.raises(ValueError):
            LoggerConfig(flush_interval_s=0.0)

    def test_invalid_buffer_size(self) -> None:
        with pytest.raises(ValueError):
            LoggerConfig(buffer_size=0)



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True

"""
OrderbookLevel data structures and construction utilities.

Provides C-level factory functions for creating OrderbookLevel structs and
Python-facing wrapper classes (PyOrderbookLevel, PyOrderbookLevels) for level
management. Includes multiple input formats: lists, numpy arrays, with optional
pre-computed ticks/lots for performance optimization.
"""
from __future__ import annotations

from typing import Self

from libc.stdint cimport uint64_t as u64
from libc.stdlib cimport free, malloc
from libc.string cimport memcpy

from .level cimport OrderbookLevel, OrderbookLevels

from .helpers cimport (
    convert_price_to_tick,
    convert_size_to_lot,
)

cdef inline OrderbookLevel create_orderbook_level(double price, double size, u64 norders=1) noexcept nogil:
    """Create an OrderbookLevel struct without pre-computing ticks/lots."""
    cdef OrderbookLevel level = OrderbookLevel()
    level.price = price
    level.size = size
    level.norders = norders
    level.ticks = 0
    level.lots = 0
    level.__padding1 = 0
    level.__padding2 = 0
    level.__padding3 = 0
    return level

cdef inline OrderbookLevel create_orderbook_level_with_ticks_and_lots(
    double price,
    double size,
    double tick_size,
    double lot_size,
    u64 norders=1,
) noexcept nogil:
    """Create an OrderbookLevel struct with pre-computed ticks and lots."""
    cdef OrderbookLevel level = OrderbookLevel()
    level.price = price
    level.size = size
    level.norders = norders
    level.ticks = convert_price_to_tick(price, tick_size)
    level.lots = convert_size_to_lot(size, lot_size)
    level.__padding1 = 0
    level.__padding2 = 0
    level.__padding3 = 0
    return level

cdef inline OrderbookLevels create_orderbook_levels(u64 num_levels, OrderbookLevel* levels) noexcept nogil:
    """Create an OrderbookLevels struct from a pointer and count."""
    cdef OrderbookLevels levels_struct = OrderbookLevels()
    levels_struct.num_levels = num_levels
    levels_struct.levels = levels
    return levels_struct


cdef inline void free_orderbook_levels(OrderbookLevels* levels) noexcept nogil:
    """Free the memory allocated for OrderbookLevels.levels if not NULL."""
    if levels != NULL and levels.levels != NULL:
        free(<void*> levels.levels)
        levels.levels = NULL
        levels.num_levels = 0


cdef class PyOrderbookLevel:
    """Python wrapper around the C OrderbookLevel struct."""
    def __cinit__(
        self, 
        double price, 
        double size, 
        u64 norders=1, 
        u64 ticks=0, 
        u64 lots=0, 
        bint verify_values=True
    ) -> None:
        """
        Create a new PyOrderbookLevel instance.

        Args:
            price: The price of the orderbook level.
            size: The size of the orderbook level.
            norders: The number of orders at the orderbook level.
            ticks: The ticks of the orderbook level.
            lots: The lots of the orderbook level.
            verify_values: Whether to verify the values of the orderbook level.
        """
        if verify_values:
            if price <= 0.0:
                raise ValueError(f"Invalid price; expected >0 but got {price}")
            if size < 0.0:
                raise ValueError(f"Invalid size; expected >=0 but got {size}")
            if norders < 0:
                raise ValueError(f"Invalid norders; expected >=0 but got {norders}")
            if ticks < 0:
                raise ValueError(f"Invalid ticks; expected >=0 but got {ticks}")
            if lots < 0:
                raise ValueError(f"Invalid lots; expected >=0 but got {lots}")

        # Create the struct directly (can't use create_orderbook_level since it doesn't accept ticks/lots)
        self._level.price = price
        self._level.size = size
        self._level.norders = norders
        self._level.ticks = ticks
        self._level.lots = lots
        self._level.__padding1 = 0
        self._level.__padding2 = 0
        self._level.__padding3 = 0

    @staticmethod
    def with_ticks_and_lots(
        double price, 
        double size, 
        double tick_size, 
        double lot_size, 
        u64 norders=1, 
        bint verify_values=True
    ) -> Self:
        """Create a new PyOrderbookLevel instance with ticks and lots."""
        cdef PyOrderbookLevel wrapper = PyOrderbookLevel(
            price=price,
            size=size,
            norders=norders,
            ticks=convert_price_to_tick(price, tick_size),
            lots=convert_size_to_lot(size, lot_size),
            verify_values=verify_values,
        )
        return wrapper

    @staticmethod
    cdef PyOrderbookLevel from_struct(OrderbookLevel level):
        """Create a PyOrderbookLevel from a C OrderbookLevel struct.
        
        Args:
            level: The C OrderbookLevel struct to wrap.
            
        Returns:
            A new PyOrderbookLevel instance wrapping the struct.
        """
        cdef PyOrderbookLevel wrapper = PyOrderbookLevel.__new__(PyOrderbookLevel)
        wrapper._level = level
        return wrapper

    cdef OrderbookLevel to_c_struct(self):
        """Return the underlying C OrderbookLevel struct (Cython only)."""
        return self._level

    @property
    def price(self) -> float:
        """Get the price of this level."""
        return self._level.price

    @property
    def size(self) -> float:
        """Get the size of this level."""
        return self._level.size

    @property
    def norders(self) -> int:
        """Get the number of orders at this level."""
        return self._level.norders

    @property
    def ticks(self) -> int:
        """Get the price in ticks."""
        return self._level.ticks

    @property
    def lots(self) -> int:
        """Get the size in lots."""
        return self._level.lots

    def __repr__(self) -> str:
        """Return a print-friendly representation of the PyOrderbookLevel instance."""
        return f"PyOrderbookLevel(price={self._level.price}, size={self._level.size}, norders={self._level.norders}, ticks={self._level.ticks}, lots={self._level.lots})"


cdef class PyOrderbookLevels:
    """Python wrapper around the C OrderbookLevels struct."""
    
    @staticmethod
    cdef PyOrderbookLevels _create(u64 num_levels, OrderbookLevel* levels):
        """Internal factory for creating PyOrderbookLevels from C types."""
        cdef PyOrderbookLevels obj = PyOrderbookLevels.__new__(PyOrderbookLevels)
        obj._levels = create_orderbook_levels(num_levels, levels)
        return obj
    
    def __cinit__(self) -> None:
        """Initialize an empty PyOrderbookLevels instance."""
        # Fields initialized by _create or factory methods
        self._levels.num_levels = 0
        self._levels.levels = NULL

    def __dealloc__(self) -> None:
        """Deallocate the PyOrderbookLevels instance."""
        free_orderbook_levels(&self._levels)
        # Note: _levels is a struct, free_orderbook_levels already resets fields

    @staticmethod
    cdef PyOrderbookLevels from_ptr(OrderbookLevel* levels_ptr, u64 num_levels):
        """Create a new PyOrderbookLevels instance from a pointer and count (Cython only)."""
        if levels_ptr == NULL:
            raise ValueError("Invalid levels_ptr; expected non-null")
        if num_levels <= 0:
            raise ValueError(f"Invalid num_levels; expected >0 but got {num_levels}")

        cdef u64 size = sizeof(OrderbookLevel) * num_levels
        cdef OrderbookLevel* new_levels = <OrderbookLevel*> malloc(size)

        if new_levels == NULL:
            raise MemoryError("Failed to allocate memory for new levels")

        memcpy(new_levels, levels_ptr, size)

        return PyOrderbookLevels._create(num_levels, new_levels)

    @staticmethod
    def from_list(list[double] prices, list[double] sizes, list[u64] norders=None, bint verify_values=True) -> Self:
        """Create a new PyOrderbookLevels instance from a list of prices, sizes, and norders."""
        if verify_values:
            len_prices = len(prices)
            len_sizes = len(sizes)
            len_norders = len(norders) if norders is not None else 0
            if len_prices <= 0:
                raise ValueError(f"Invalid prices; expected >0 but got {len_prices}")
            if len_sizes != len_prices:
                raise ValueError(f"Mismatched lengths; expected same length for prices and sizes but got {len_prices} and {len_sizes}")
            if norders is not None and len_norders != len_prices:
                raise ValueError(f"Mismatched lengths; expected same length for prices and norders but got {len_prices} and {len_norders}")

        cdef:
            u64 i
            u64 size = sizeof(OrderbookLevel) * len_prices
            bint use_default = norders is None or len_norders == 0
            OrderbookLevel* new_levels = <OrderbookLevel*> malloc(size)

        if new_levels == NULL:
            raise MemoryError("Failed to allocate memory for new levels")

        try:
            for i in range(len_prices):
                new_levels[i] = create_orderbook_level(
                    price=prices[i],
                    size=sizes[i],
                    norders=norders[i] if not use_default else 1,
                )
        except:
            free(new_levels)
            raise

        return PyOrderbookLevels._create(<u64>len_prices, new_levels)

    @staticmethod
    def from_list_with_ticks_and_lots(
        list[double] prices,
        list[double] sizes,
        list[u64] norders,
        double tick_size,
        double lot_size,
        bint verify_values=True,
    ) -> Self:
        """Create a new PyOrderbookLevels instance from lists with pre-computed ticks and lots."""
        cdef Py_ssize_t num = len(prices)
        if verify_values:
            if num <= 0:
                raise ValueError(f"Invalid prices; expected >0 but got {num}")
            if len(sizes) != num:
                raise ValueError(f"Mismatched lengths; expected same length for prices and sizes but got {num} and {len(sizes)}")
            if len(norders) != num:
                raise ValueError(f"Mismatched lengths; expected same length for prices and norders but got {num} and {len(norders)}")

        cdef:
            u64 i
            u64 size = sizeof(OrderbookLevel) * num
            OrderbookLevel* new_levels = <OrderbookLevel*> malloc(size)

        if new_levels == NULL:
            raise MemoryError("Failed to allocate memory for new levels")

        try:
            for i in range(num):
                new_levels[i] = create_orderbook_level_with_ticks_and_lots(
                    prices[i],
                    sizes[i],
                    tick_size,
                    lot_size,
                    norders[i],
                )
        except:
            free(new_levels)
            raise

        return PyOrderbookLevels._create(<u64>num, new_levels)

    @staticmethod
    def from_numpy(double[:] prices, double[:] sizes, u64[:] norders=None, bint verify_values=True) -> Self:
        """Create a new PyOrderbookLevels instance from numpy arrays."""
        cdef Py_ssize_t num = prices.shape[0]
        cdef bint use_default = norders is None

        if verify_values:
            if num <= 0:
                raise ValueError(f"Invalid prices; expected >0 but got {num}")
            if sizes.shape[0] != num:
                raise ValueError(f"Mismatched lengths; expected same length for prices and sizes but got {num} and {sizes.shape[0]}")
            if not use_default and norders.shape[0] != num:
                raise ValueError(f"Mismatched lengths; expected same length for prices and norders but got {num} and {norders.shape[0]}")

        cdef:
            Py_ssize_t i
            u64 size = sizeof(OrderbookLevel) * num
            OrderbookLevel* new_levels = <OrderbookLevel*> malloc(size)

        if new_levels == NULL:
            raise MemoryError("Failed to allocate memory for new levels")

        try:
            for i in range(num):
                new_levels[i] = create_orderbook_level(
                    price=prices[i],
                    size=sizes[i],
                    norders=norders[i] if not use_default else 1,
                )
        except:
            free(new_levels)
            raise

        return PyOrderbookLevels._create(<u64>num, new_levels)

    @staticmethod
    def from_numpy_with_ticks_and_lots(
        double[:] prices,
        double[:] sizes,
        u64[:] norders,
        double tick_size,
        double lot_size,
        bint verify_values=True,
    ) -> Self:
        """Create a new PyOrderbookLevels instance from numpy arrays with pre-computed ticks and lots."""
        cdef Py_ssize_t num = prices.shape[0]

        if verify_values:
            if num <= 0:
                raise ValueError(f"Invalid prices; expected >0 but got {num}")
            if sizes.shape[0] != num:
                raise ValueError(f"Mismatched lengths; expected same length for prices and sizes but got {num} and {sizes.shape[0]}")
            if norders.shape[0] != num:
                raise ValueError(f"Mismatched lengths; expected same length for prices and norders but got {num} and {norders.shape[0]}")

        cdef:
            Py_ssize_t i
            u64 size = sizeof(OrderbookLevel) * num
            OrderbookLevel* new_levels = <OrderbookLevel*> malloc(size)

        if new_levels == NULL:
            raise MemoryError("Failed to allocate memory for new levels")

        try:
            for i in range(num):
                new_levels[i] = create_orderbook_level_with_ticks_and_lots(
                    prices[i],
                    sizes[i],
                    tick_size,
                    lot_size,
                    norders[i],
                )
        except:
            free(new_levels)
            raise

        return PyOrderbookLevels._create(<u64>num, new_levels)

    cdef OrderbookLevels to_c_struct(self):
        """Return the underlying C OrderbookLevels struct."""
        return self._levels
